/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007 2008 2009 2010 2011 2012 2013 Michael Cornelison
   Source URL: http://kornelix.com/fotoxx
   Contact: kornelix@gmx.de
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************/

#define EX extern                                                          //  enable extern declarations
#include "fotoxx.h"

/**************************************************************************
   Fotoxx image edit - Effects menu functions
***************************************************************************/


//  image color-depth reduction

int      colordep_depth = 16;                                              //  bits per RGB color

editfunc    EFcolordep;


void m_colordep(GtkWidget *, cchar *)
{
   int    colordep_dialog_event(zdialog *zd, cchar *event);
   void * colordep_thread(void *);

   cchar  *colmess = ZTX("Set color depth to 1-16 bits");
   
   F1_help_topic = "color_depth";

   EFcolordep.funcname = "color-depth";
   EFcolordep.Fprev = 1;                                                   //  use preview
   EFcolordep.Farea = 2;                                                   //  select area usable
   EFcolordep.threadfunc = colordep_thread;                                //  thread function
   if (! edit_setup(EFcolordep)) return;                                   //  setup edit

   zdialog *zd = zdialog_new(ZTX("Set Color Depth"),mWin,Bdone,Bcancel,null);
   EFcolordep.zd = zd;

   zdialog_add_widget(zd,"label","lab1","dialog",colmess,"space=5");
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zd,"spin","colors","hb1","1|16|1|16","space=5");

   zdialog_run(zd,colordep_dialog_event,"-10/20");                         //  run dialog, parallel
   
   colordep_depth = 16;
   return;
}


//  dialog event and completion callback function

int colordep_dialog_event(zdialog *zd, cchar *event)
{
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 0;
   }

   if (strEqu(event,"apply")) {                                            //  from f_save()                v.13.04
      edit_fullsize();
      return 1;
   }
   
   if (strEqu(event,"undo")) edit_undo();
   if (strEqu(event,"redo")) edit_redo();
   if (strEqu(event,"blendwidth")) signal_thread();

   if (strEqu(event,"colors")) {
      zdialog_fetch(zd,"colors",colordep_depth);
      signal_thread();
   }
   
   return 0;
}


//  image color depth thread function

void * colordep_thread(void *)
{
   int         ii, px, py, rgb, dist = 0;
   uint16      m1, m2, val1, val3;
   float       *pix1, *pix3;
   float       fmag, f1, f2;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
   
      m1 = 0xFFFF << (16 - colordep_depth);                                //  5 > 1111100000000000
      m2 = 0x8000 >> colordep_depth;                                       //  5 > 0000010000000000
      
      fmag = 65535.0 / m1;

      for (py = 0; py < E3pxm->hh; py++)
      for (px = 0; px < E3pxm->ww; px++)
      {
         if (sa_stat == 3) {                                               //  select area active
            ii = py * Fpxb->ww + px;
            dist = sa_pixmap[ii];                                          //  distance from edge
            if (! dist) continue;                                          //  outside pixel
         }

         pix1 = PXMpix(E1pxm,px,py);                                       //  input pixel
         pix3 = PXMpix(E3pxm,px,py);                                       //  output pixel
         
         for (rgb = 0; rgb < 3; rgb++)
         {
            val1 = 256 * pix1[rgb];                                        //  16 bit integer <= 65535
            if (val1 < m1) val3 = (val1 + m2) & m1;
            else val3 = m1;
            val3 = val3 * fmag;

            if (sa_stat == 3 && dist < sa_blend) {                         //  select area is active,
               f2 = 1.0 * dist / sa_blend;                                 //    blend changes over sa_blend
               f1 = 1.0 - f2;
               val3 = int(f1 * val1 + f2 * val3);
            }

            pix3[rgb] = val3 / 256;
         }
      }

      CEF->Fmods++; 
      CEF->Fsaved = 0;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************/

//  sketch menu function - convert photo to simulated pencil sketch

namespace sketch_names 
{
   editfunc    EFsketch;                                                   //  edit function data
   float       Bweight = 0.5;                                              //  brightness weight 0-1
   int         Brthresh = 255;                                             //  brightness threshold 0-255
   float       Cweight = 0.5;                                              //  contrast weight 0-1
   int         cliplev = 255;                                              //  clipping level 0-255
   int         algorithm = 1;                                              //  which algorithm to use
   int         negative = 0;

   int         ww, hh, cc;                                                 //  image dimensions
   uint8       *britness;                                                  //  maps pixel brightness 0-255
   uint8       *CLcont;                                                    //  maps pixel color contrast 0-255
   uint8       *monopix;                                                   //  output pixel map 0-255
   uint8       *fclip;                                                     //  output clipping flag 0/1
}


void m_sketch(GtkWidget *, const char *)                                   //  overhauled    v.13.08
{
   using namespace sketch_names;

   int    sketch_dialog_event(zdialog* zd, const char *event);
   void * sketch_thread(void *);

   int      ii, px, py;
   float    *pix0, *pix1, *pix2, *pix3, *pix4;
   float    fred, fgreen, fblue;
   float    f1, f2, f256 = 1.0 / 256.0;

   cchar    *title = ZTX("Convert to Sketch");

   F1_help_topic = "sketch";
   
   EFsketch.funcname = "sketch";                                           //  function name
   EFsketch.Farea = 2;                                                     //  select area usable
   EFsketch.threadfunc = sketch_thread;                                    //  thread function

   if (! edit_setup(EFsketch)) return;                                     //  setup edit

   ww = Fpxb->ww;
   hh = Fpxb->hh;
   cc = ww * hh;

   britness = (uint8 *) zmalloc(cc,"sketch");
   CLcont = (uint8 *) zmalloc(cc,"sketch");
   monopix = (uint8 *) zmalloc(cc,"sketch");   
   fclip = (uint8 *) zmalloc(cc,"sketch");   

   for (py = 1; py < hh-1; py++)                                           //  precalculate pixel brightness   v.13.08
   for (px = 1; px < ww-1; px++)
   {
      pix0 = PXMpix(E1pxm,px,py);
      ii = py * ww + px;
      britness[ii] = 0.333 * (pix0[0] + pix0[1] + pix0[2]);                //  brightness 0-255  (pixbright() nodiff)
   }

   for (py = 1; py < hh-1; py++)                                           //  precalculate pixel contrast     v.13.08
   for (px = 1; px < ww-1; px++)
   {
      pix0 = PXMpix(E1pxm,px,py);
      
      pix1 = PXMpix(E1pxm,px-1,py);                                        //  get pixel contrast
      pix2 = PXMpix(E1pxm,px+1,py);
      pix3 = PXMpix(E1pxm,px,py-1);
      pix4 = PXMpix(E1pxm,px,py+1);
      
      fred = f256 * (fabsf(pix1[0] - pix2[0]));                            //  horizontal contrast
      fgreen = f256 * (fabsf(pix1[1] - pix2[1]));
      fblue = f256 * (fabsf(pix1[2] - pix2[2]));
      f1 = (1.0 - fred) * (1.0 - fgreen) * (1.0 - fblue);                  //  1 = perfect match (0 contrast)

      fred = f256 * (fabsf(pix3[0] - pix4[0]));                            //  vertical contrast
      fgreen = f256 * (fabsf(pix3[1] - pix4[1]));
      fblue = f256 * (fabsf(pix3[2] - pix4[2]));
      f2 = (1.0 - fred) * (1.0 - fgreen) * (1.0 - fblue);                  //  1 = perfect match
      
      ii = py * ww + px;
      CLcont[ii] = 255.0 * (1.0 - f1 * f2);                                //  pixel color contrast 0-255
   }

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);               //  sketch dialog
   EFsketch.zd = zd;

/***
        _____________________________________
       |        Convert to Sketch            |
       |                                     |
       |  Brightness ===========[]========== |
       |  Threshold ======[]================ |
       |  Contrast =========[]============== |
       |  Clip Level ================[]===== |
       |  Algorithm  (o) #1  (o) #2          |
       |  [x] Negative                       |
       |                                     |
       |                    [done] [cancel]  |
       |_____________________________________|

***/
   
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=2");
   zdialog_add_widget(zd,"label","labbrite","hb1",Bbrightness,"space=5");
   zdialog_add_widget(zd,"hscale","Bweight","hb1","0.0|1.0|0.005|0.5","expand|space=3");

   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=2");
   zdialog_add_widget(zd,"label","laBrthresh","hb2",Bthresh,"space=5");
   zdialog_add_widget(zd,"hscale","Brthresh","hb2","0|255|1|255","expand|space=3");

   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=2");
   zdialog_add_widget(zd,"label","labcon","hb3",Bcontrast,"space=5");
   zdialog_add_widget(zd,"hscale","Cweight","hb3","0.0|1.0|0.005|0.5","expand|space=3");

   zdialog_add_widget(zd,"hbox","hb4","dialog",0,"space=2");
   zdialog_add_widget(zd,"label","labclip","hb4",ZTX("Clip Level"),"space=5");
   zdialog_add_widget(zd,"hscale","cliplev","hb4","0|255|1|0","expand|space=3");

   zdialog_add_widget(zd,"hbox","hb5","dialog",0,"space=2");
   zdialog_add_widget(zd,"label","labalg","hb5",ZTX("Algorithm"),"space=5");
   zdialog_add_widget(zd,"radio","algorithm1","hb5","#1","space=5");
   zdialog_add_widget(zd,"radio","algorithm2","hb5","#2","space=5");

   zdialog_add_widget(zd,"hbox","hb6","dialog",0,"space=2");
   zdialog_add_widget(zd,"check","negative","hb6",Bnegative,"space=2");
   
   Bweight = 0.5;                                                          //  initial values                  v.13.08
   Brthresh = 255;
   Cweight = 0.5;
   cliplev = 255;
   algorithm = 1;
   negative = 0;
   zdialog_stuff(zd,"algorithm1",1);
   zdialog_stuff(zd,"negative",0);

   zdialog_resize(zd,250,0);
   zdialog_run(zd,sketch_dialog_event,"save");                             //  run dialog - parallel

   signal_thread();
   return;
}


//  sketch dialog event and completion function

int sketch_dialog_event(zdialog *zd, const char *event)
{
   using namespace sketch_names;
   
   int      ii;

   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();                                     //  done
      else edit_cancel();                                                  //  cancel or destroy
      zfree(britness);
      zfree(CLcont);
      zfree(monopix);
      zfree(fclip);
      return 0;
   }
   
   if (strEqu(event,"Bweight")) {
      zdialog_fetch(zd,"Bweight",Bweight);                                 //  revised brightness weight
      signal_thread();                                                     //  trigger update thread
   }

   if (strEqu(event,"Brthresh")) {
      zdialog_fetch(zd,"Brthresh",Brthresh);                               //  brightness threshold
      signal_thread();
   }

   if (strEqu(event,"Cweight")) {
      zdialog_fetch(zd,"Cweight",Cweight);                                 //  contrast weight
      signal_thread();
   }
   
   if (strEqu(event,"cliplev")) {                                          //  output clip level
      zdialog_fetch(zd,"cliplev",cliplev);
      cliplev = 255 - cliplev;                                             //  scale is reversed
      signal_thread();
   }
   
   if (strstr(event,"algorithm")) {                                        //  algorithm to use
      zdialog_fetch(zd,"algorithm1",ii);
      if (ii == 1) algorithm = 1;
      else algorithm = 2;
      signal_thread();
   }
   
   if (strEqu(event,"negative")) {
      zdialog_fetch(zd,"negative",negative);                               //  revised negative
      signal_thread();
   }

   return 1;
}


//  thread function - multiple working threads to update image

void * sketch_thread(void *)
{
   using namespace sketch_names;

   void sketch_algorithm1();
   void sketch_algorithm2();
   void sketch_finish();

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      if (algorithm == 1) sketch_algorithm1();
      if (algorithm == 2) sketch_algorithm2();
      sketch_finish();

      CEF->Fmods = 1;                                                      //  image modified
      CEF->Fsaved = 0;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop warning
}


//  algorithm 1 

void sketch_algorithm1()
{
   using namespace sketch_names;

   int         px, py, qx, qy;
   int         ii, jj, kk, dist;
   int         br0, br1, br2, br3, move;
   float       bright, cont;

   for (py = 0; py < hh; py++)                                             //  convert to monocolor image
   for (px = 0; px < ww; px++)                                             //    with brightness range 0-255
   {
      ii = py * ww + px;

      if (sa_stat == 3) {                                                  //  select area active
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel outside area
      }
         
      bright = britness[ii];
      if (bright > Brthresh) bright = 255;
      cont = CLcont[ii];                                                   //  contrast 0-255
      bright = Bweight * bright - Cweight * cont;
      if (bright < 0) bright = 0;
      monopix[ii] = bright;
   }

   for (br0 = 5; br0 <= 255; br0 += 10)                                    //  brightness threshold, 5 ... 255
   {
      for (py = 1; py < hh-1; py++)                                        //  loop all image pixels
      for (px = 1; px < ww-1; px++)                                        //  (interior pixels only)
      {
         ii = py * ww + px;

         if (sa_stat == 3) {                                               //  select area active
            dist = sa_pixmap[ii];                                          //  distance from edge
            if (! dist) continue;                                          //  pixel outside area
         }
         
         br1 = monopix[ii];                                                //  pixel brightness
         if (br1 == 0) continue;                                           //  skip black pixel
         if (br1 > br0) continue;                                          //  skip pixel brighter than br0
         
         for (qy = py-1; qy <= py+1; qy++)                                 //  loop pixels within 1 of target pixel
         for (qx = px-1; qx <= px+1; qx++)                                 //  test if target pixel is darkest > 0
         {
            jj = qy * ww + qx;
            br2 = monopix[jj];
            if (br2 == 0) continue;                                        //  ignore black pixels
            if (br2 < br1) goto nextpixel;                                 //  target pixel not darkest
         }

         br2 = 0;
         kk = -1;

         for (qy = py-1; qy <= py+1; qy++)                                 //  loop pixels within 1 of target pixel
         for (qx = px-1; qx <= px+1; qx++)                                 //  find brightest pixel < 255
         {
            jj = qy * ww + qx;
            br3 = monopix[jj];
            if (br3 < 255 && br3 > br2) {
               br2 = br3;
               kk = jj;
            }
         }
         
         if (kk < 0) continue;

         move = br1;                                                       //  move brightness from target pixel,
         if (255 - br2 < move) move = 255 - br2;                           //    as much as possible 
         monopix[ii] -= move;
         monopix[kk] += move;
         
         nextpixel: continue;
      }
   }

   return;                                                                 //  not executed, avoid gcc warning
}


//  algorithm 2

void sketch_algorithm2()
{
   using namespace sketch_names;

   int         px, py, qx, qy;
   int         ii, jj, kk, dist;
   int         br0, br1, br2, br3, move;
   int         bright, cont, brcon;
   
   for (py = 0; py < hh; py++)                                             //  convert to monocolor image
   for (px = 0; px < ww; px++)                                             //    with brightness range 0-255
   {
      ii = py * ww + px;

      if (sa_stat == 3) {                                                  //  select area active
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel outside area
      }

      bright = Bweight * britness[ii];                                     //  brightness 0-255
      if (bright > Brthresh) bright = 255;
      cont = CLcont[ii];                                                   //  contrast 0-255

      brcon = (1.0 - Cweight) * bright - Cweight * cont;                   //  255 = max. brightness + min. contrast
      if (brcon < 0) brcon = 0;
      if (brcon > 0) brcon = 255;                                          //  pixels are black/white
      monopix[ii] = brcon;                                                 //  0-255
   }
   
   for (br0 = 5; br0 <= 255; br0 += 50)                                    //  loop brightness threshold
   {
      for (py = 1; py < hh-1; py++)                                        //  loop image interior pixels
      for (px = 1; px < ww-1; px++)
      {
         ii = py * ww + px;

         if (sa_stat == 3) {                                               //  select area active
            dist = sa_pixmap[ii];                                          //  distance from edge
            if (! dist) continue;                                          //  pixel outside area
         }
         
         br1 = monopix[ii];
         if (br1 > br0 || br1 == 0) continue;                              //  skip target pixel > threshold or black
         
         for (qy = py-1; qy <= py+1; qy++)                                 //  loop pixels within 1 of target pixel
         for (qx = px-1; qx <= px+1; qx++)
         {
            jj = qy * ww + qx;
            br2 = monopix[jj];
            if (br2 == 0) continue;                                        //  ignore black pixels
            if (br2 < br1) goto nextpixel;                                 //  target pixel not darkest
         }

         br2 = 0;
         kk = -1;

         for (qy = py-1; qy <= py+1; qy++)                                 //  loop pixels within 1 of target pixel
         for (qx = px-1; qx <= px+1; qx++)
         {
            jj = qy * ww + qx;
            br3 = monopix[jj];
            if (br3 < 255 && br3 > br2) {                                  //  find brightest < 255
               br2 = br3;
               kk = jj;
            }
         }
         
         if (kk < 0) continue;

         move = br1;                                                       //  move darkness to target pixel,
         if (255 - br2 < move) move = 255 - br2;                           //    as much as possible 
         br1 -= move;
         br2 += move;
         monopix[ii] = br1;
         monopix[kk] = br2;
            
         nextpixel: ;
      }
   }

   return;                                                                 //  not executed, avoid gcc warning
}


//  convert monopix[*] to final image

void sketch_finish()
{
   using namespace sketch_names;

   int         px, py, qx, qy;
   int         ii, jj, dist;
   int         br1, brsum;
   float       *pix1, *pix3, f1, f2;

   memset(fclip,0,cc);
   
   for (py = 1; py < hh-1; py++)                                           //  loop all pixels
   for (px = 1; px < ww-1; px++)                                           //  (interior pixels only)
   {
      ii = py * ww + px;                 

      if (sa_stat == 3) {                                                  //  select area active
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel outside area
      }

      brsum = 0;

      for (qy = py-1; qy <= py+1; qy++)                                    //  loop pixels within 1 of target pixel
      for (qx = px-1; qx <= px+1; qx++)
      {
         jj = qy * ww + qx;                                                //  sum brightness
         brsum += monopix[jj];
      }
      
      brsum = brsum / 9;                                                   //  average brightness

      if (brsum > cliplev) fclip[ii] = 1;                                  //  reduce isolated dark pixels
   }
   
   for (ii = 0; ii < cc; ii++)
      if (fclip[ii]) monopix[ii] = 255;

   for (py = 0; py < hh; py++)                                             //  loop all pixels
   for (px = 0; px < ww; px++)
   {
      ii = py * ww + px;

      br1 = monopix[ii];                                                   //  input pixel (monocolor brightness)
      if (negative) br1 = 255 - br1;                                       //  v.13.08
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel

      if (sa_stat == 3) {                                                  //  select area active
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel outside area
         if (dist < sa_blend) {
            pix1 = PXMpix(E1pxm,px,py);                                    //  input pixel (blend area)
            f1 = 1.0 * dist / sa_blend;
            f2 = 1.0 - f1;
            pix3[0] = f1 * br1 + f2 * pix1[0];                             //  blend monocolor brightness
            pix3[1] = f1 * br1 + f2 * pix1[1];                             //    and original image 
            pix3[2] = f1 * br1 + f2 * pix1[2];
         }
         else                                                              //  beyond blend area
            pix3[0] = pix3[1] = pix3[2] = br1;                             //  output pixel is monocolor brightness
         continue;
      }
      
      pix3[0] = pix3[1] = pix3[2] = br1;                                   //  output pixel is monocolor brightness
   }

   return;
}


/**************************************************************************/

//  make an image outline drawing from edge pixels
//  superimpose drawing and image and vary brightness of each

namespace drawing_names
{
   float       outline_thresh;                                             //  contrast threshold to make a line
   float       outline_width;                                              //  drawn line width
   float       image_briteness;                                            //  background image brightness
   int         blackwhite, negative;                                       //  flags
   editfunc    EFdrawing;
}

void m_drawing(GtkWidget *, cchar *)
{
   using namespace drawing_names;

   int    drawing_dialog_event(zdialog* zd, cchar *event);
   void * drawing_thread(void *);

   cchar  *title = ZTX("Simulate Drawing");

   F1_help_topic = "drawing";
   
   EFdrawing.funcname = "drawing";
   EFdrawing.Farea = 2;                                                    //  select area usable
   EFdrawing.threadfunc = drawing_thread;                                  //  thread function
   if (! edit_setup(EFdrawing)) return;                                    //  setup edit

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);
   EFdrawing.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0);
   zdialog_add_widget(zd,"label","lab1","hb1",ZTX("outline threshold"),"space=5");
   zdialog_add_widget(zd,"hscale","olth","hb1","0|100|1|90","expand|space=5");   
   zdialog_add_widget(zd,"hbox","hb2","dialog",0);
   zdialog_add_widget(zd,"label","lab2","hb2",ZTX("outline width"),"space=5");
   zdialog_add_widget(zd,"hscale","olww","hb2","0|100|1|50","expand|space=5");
   zdialog_add_widget(zd,"hbox","hb3","dialog",0);
   zdialog_add_widget(zd,"label","lab3","hb3",ZTX("image brightness"),"space=5");
   zdialog_add_widget(zd,"hscale","imbr","hb3","0|100|1|10","expand|space=5");
   zdialog_add_widget(zd,"hbox","hb4","dialog",0);
   zdialog_add_widget(zd,"check","blackwhite","hb4",ZTX("black/white"),"space=5");
   zdialog_add_widget(zd,"label","space","hb4",0,"space=10");
   zdialog_add_widget(zd,"check","negative","hb4",Bnegative);
   
   outline_thresh = 90;
   outline_width = 50;
   image_briteness = 10;
   blackwhite = negative = 0;

   zdialog_stuff(zd,"blackwhite",0);
   zdialog_stuff(zd,"negative",0);

   zdialog_resize(zd,300,0);
   zdialog_run(zd,drawing_dialog_event,"-10/20");                          //  run dialog, parallel

   signal_thread();
   return;
}


//  dialog event and completion callback function

int drawing_dialog_event(zdialog *zd, cchar *event)                        //  dialog event function
{
   using namespace drawing_names;

   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();                                     //  done
      else edit_cancel();                                                  //  cancel or destroy
      PXM_free(E8pxm);
      PXM_free(E9pxm);
      return 0;
   }

   if (strEqu(event,"olth")) {
      zdialog_fetch(zd,"olth",outline_thresh);                             //  get outline threshold 0-100
      signal_thread();
   }

   if (strEqu(event,"olww")) {
      zdialog_fetch(zd,"olww",outline_width);                              //  get outline width 0-100
      signal_thread();
   }
   
   if (strEqu(event,"imbr")) {
      zdialog_fetch(zd,"imbr",image_briteness);                            //  get image brightness 0-100
      signal_thread();
   }
   
   if (strEqu(event,"blackwhite")) {                                       //  v.13.07
      zdialog_fetch(zd,"blackwhite",blackwhite);
      signal_thread();
   }
   
   if (strEqu(event,"negative")) {                                         //  v.13.07
      zdialog_fetch(zd,"negative",negative);
      signal_thread();
   }
   
   if (strEqu(event,"undo")) edit_undo();
   if (strEqu(event,"redo")) edit_redo();
   if (strEqu(event,"blendwidth")) signal_thread();

   return 1;
}


//  thread function to update image

void * drawing_thread(void *)
{
   using namespace drawing_names;

   void * drawing_wthread(void *arg);
   
   int         px, py, ww, hh;
   float       olww, red3, green3, blue3;
   float       red3h, red3v, green3h, green3v, blue3h, blue3v;
   float       *pix8, *pix9;
   float       *pixa, *pixb, *pixc, *pixd, *pixe, *pixf, *pixg, *pixh;

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      olww = 0.01 * outline_width;                                         //  outline width, 0 - 1.0
      olww = 1.0 - olww;                                                   //  1.0 - 0
      olww = 0.8 * olww + 0.2;                                             //  1.0 - 0.2

      ww = Fpxb->ww * olww + 0.5;                                          //  create smaller outline image
      hh = Fpxb->hh * olww + 0.5;
      
      if (! E9pxm || ww != E9pxm->ww)                                      //  initial or changed outline brightness
      {
         PXM_free(E8pxm);
         PXM_free(E9pxm);

         E8pxm = PXM_rescale(E1pxm,ww,hh);
         E9pxm = PXM_copy(E8pxm);

         for (py = 1; py < hh-1; py++)
         for (px = 1; px < ww-1; px++)
         {
            pix8 = PXMpix(E8pxm,px,py);                                    //  input pixel
            pix9 = PXMpix(E9pxm,px,py);                                    //  output pixel
            
            pixa = pix8-3*ww-3;                                            //  8 neighboring pixels are used
            pixb = pix8-3*ww;                                              //    to get edge brightness using
            pixc = pix8-3*ww+3;                                            //       a Sobel filter
            pixd = pix8-3;
            pixe = pix8+3;
            pixf = pix8+3*ww-3;
            pixg = pix8+3*ww;
            pixh = pix8+3*ww+3;
            
            red3h = -pixa[0] - 2 * pixb[0] - pixc[0] + pixf[0] + 2 * pixg[0] + pixh[0];
            red3v = -pixa[0] - 2 * pixd[0] - pixf[0] + pixc[0] + 2 * pixe[0] + pixh[0];
            green3h = -pixa[1] - 2 * pixb[1] - pixc[1] + pixf[1] + 2 * pixg[1] + pixh[1];
            green3v = -pixa[1] - 2 * pixd[1] - pixf[1] + pixc[1] + 2 * pixe[1] + pixh[1];
            blue3h = -pixa[2] - 2 * pixb[2] - pixc[2] + pixf[2] + 2 * pixg[2] + pixh[2];
            blue3v = -pixa[2] - 2 * pixd[2] - pixf[2] + pixc[2] + 2 * pixe[2] + pixh[2];
            
            red3 = (fabsf(red3h) + fabsf(red3v)) / 2;                      //  average vertical and horizontal brightness
            green3 = (fabsf(green3h) + fabsf(green3v)) / 2;
            blue3 = (fabsf(blue3h) + fabsf(blue3v)) / 2;
            
            if (red3 > 255.9) red3 = 255.9;
            if (green3 > 255.9) green3 = 255.9;
            if (blue3 > 255.9) blue3 = 255.9;

            pix9[0] = red3;
            pix9[1] = green3;
            pix9[2] = blue3;
         }
      }
      
      PXM_free(E8pxm);                                                     //  scale back to full-size
      E8pxm = PXM_rescale(E9pxm,Fpxb->ww,Fpxb->hh);

      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(drawing_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      CEF->Fmods++; 
      CEF->Fsaved = 0;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * drawing_wthread(void *arg)                                          //  worker thread function
{
   using namespace drawing_names;

   int         index = *((int *) arg);
   int         px, py, ii, dist = 0;
   float       olth, imbr, br8, dold, dnew;
   float       red1, green1, blue1, red3, green3, blue3;
   float       *pix1, *pix8, *pix3;
   
   olth = 0.01 * outline_thresh;                                           //  outline threshold, 0 - 1.0
   olth = 1.0 - olth;                                                      //                     1.0 - 0
   imbr = 0.01 * image_briteness;                                          //  image brightness,  0 - 1.0

   for (py = index+1; py < Fpxb->hh-1; py += Nwt)
   for (px = 1; px < Fpxb->ww-1; px++)
   {
      if (sa_stat == 3) {                                                  //  select area active
         ii = py * Fpxb->ww + px;
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel outside area
      }
      
      pix1 = PXMpix(E1pxm,px,py);                                          //  input image pixel
      pix8 = PXMpix(E8pxm,px,py);                                          //  input outline pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output image pixel
      
      red1 = pix1[0];                                                      //  input image pixel
      green1 = pix1[1];
      blue1 = pix1[2];
      
      br8 = pixbright(pix8);                                               //  outline brightness
      br8 = br8 / 256.0;                                                   //  scale 0 - 1.0

      if (br8 > olth) {                                                    //  > threshold, use outline pixel
         red3 = pix8[0];
         green3 = pix8[1];
         blue3 = pix8[2];
      }
      else {                                                               //  use image pixel dimmed by user control
         red3 = red1 * imbr;
         green3 = green1 * imbr;
         blue3 = blue1 * imbr;
      }
      
      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         dnew = 1.0 * dist / sa_blend;                                     //    blend changes over sa_blend
         dold = 1.0 - dnew;
         red3 = dnew * red3 + dold * red1;
         green3 = dnew * green3 + dold * green1;
         blue3 = dnew * blue3 + dold * blue1;
      }
      
      if (blackwhite)                                                      //  v.13.07
         red3 = green3 = blue3 = 0.333 * (red3 + green3 + blue3);
      
      if (negative) {                                                      //  v.13.07
         red3 = 255.9 - red3;
         green3 = 255.9 - green3;
         blue3 = 255.9 - blue3;
      }         

      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }
   
   exit_wthread();
   return 0;                                                               //  not executed, stop gcc warning
}


/**************************************************************************/

//  Graduated blur of image pixels.
//  Blur radius is zero ... max for pixel contrast max ... zero.

namespace gradblur_names 
{
   editfunc    EFgradblur;
   int         con_limit;
   float       blur_radius;
   int         gradblur_cancel;
   uint8       *pixcon;

   int         ww, hh, cc;
   int         pixseq_done[122][122];                                      //  up to blur_radius = 60
   int         pixseq_angle[1000];
   int         pixseq_dx[13000], pixseq_dy[13000];
   int         pixseq_rad[13000];
   int         max1 = 999, max2 = 12999;                                   //  for later overflow check
}


//  return a contrast value 0.0 - 1.0 for two given pixels to compare

inline float pixcontrast(float *pix1, float *pix2)
{
   float    f256 = 1.0 / 256.0;
   float    conR, conG, conB, match;

   conR = f256 * fabsf(pix1[0] - pix2[0]);                                 //  0 = perfect match
   conG = f256 * fabsf(pix1[1] - pix2[1]);
   conB = f256 * fabsf(pix1[2] - pix2[2]);
   match = (1.0 - conR) * (1.0 - conG) * (1.0 - conB);                     //  1 = perfect match

   return 1.0 - match;                                                     //  1 = high contrast (black/white)
}


void m_gradblur(GtkWidget *, cchar *)                                      //  new v.13.08
{
   using namespace gradblur_names;

   int    gradblur_dialog_event(zdialog *zd, cchar *event);
   void * gradblur_thread(void *);
   
   int      ii, px, py, dx, dy;
   float    *pix1, *pix2;
   float    contrast, maxcon;

   F1_help_topic = "grad_blur";

   EFgradblur.funcname = "gradblur";
   EFgradblur.Farea = 2;                                                   //  select area usable
   EFgradblur.threadfunc = gradblur_thread;                                //  thread function
   if (! edit_setup(EFgradblur)) return;                                   //  setup edit

   gradblur_cancel = 0;

   ww = E3pxm->ww;                                                         //  pre-calculate pixel contrasts
   hh = E3pxm->hh;
   cc = ww * hh;
   pixcon = (uint8 *) zmalloc(cc,"gradblur");                              //  pixel contrast map
   
   for (py = 1; py < hh-1; py++)                                           //  loop interior pixels
   for (px = 1; px < ww-1; px++)
   {
      pix1 = PXMpix(E1pxm,px,py);                                          //  this pixel in base image E1
      contrast = maxcon = 0.0;

      for (dx = px-1; dx <= px+1; dx++)                                    //  loop neighbor pixels
      for (dy = py-1; dy <= py+1; dy++)
      {
         pix2 = PXMpix(E1pxm,dx,dy);
         contrast = pixcontrast(pix1,pix2);                                //  contrast, 0-1
         if (contrast > maxcon) maxcon = contrast;
      }

      ii = py * ww + px;                                                   //  ii maps to (px,py)
      pixcon[ii] = 255 * maxcon;                                           //  pixel contrast, 0 to 255
   }

/***
          _____________________________
         |     Graduated Blur          |
         |                             |
         |  Contrast Limit [ 40 +-]    |
         |   Blur Radius   [ 16 +-]    |
         |                             |
         |  [apply] [done] [cancel]    |
         |_____________________________|

***/

   zdialog *zd = zdialog_new(ZTX("Graduated Blur"),mWin,Bapply,Bdone,Bcancel,null);
   EFgradblur.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","lab1","hb1",ZTX("Contrast Limit"),"space=5");
   zdialog_add_widget(zd,"spin","con_limit","hb1","1|255|1|50");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","lab5","hb2",ZTX("Blur Radius"),"space=5");
   zdialog_add_widget(zd,"spin","blur_radius","hb2","1|50|1|15");

   zdialog_restore_inputs(zd);
   zdialog_run(zd,gradblur_dialog_event,"save");                           //  run dialog - parallel

   return;
}


//  dialog event and completion callback function

int gradblur_dialog_event(zdialog * zd, cchar *event)
{
   using namespace gradblur_names;

   if (zd->zstat)
   {
      if (zd->zstat == 1)                                                  //  apply
      {
         zd->zstat = 0;                                                    //  keep dialog active
         zdialog_fetch(zd,"con_limit",con_limit);                          //  get parameters
         zdialog_fetch(zd,"blur_radius",blur_radius);
         edit_undo();                                                      //  reset image
         signal_thread();                                                  //  trigger working thread
      }

      else if (zd->zstat == 2) {                                           //  done
         edit_done();
         zfree(pixcon);
      }

      else {
         gradblur_cancel = 1;
         edit_cancel();                                                    //  cancel or destroy
         zfree(pixcon);
      }
   }

   return 1;
}


//  image gradblur thread function

void * gradblur_thread(void *)
{
   using namespace gradblur_names;

   void * gradblur_wthread(void *arg);

   int         dx, dy, adx, ady;
   int         ii, jj;
   float       rad1, rad2, angle, astep;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      rad1 = blur_radius;

      for (dy = 0; dy <= 2*rad1; dy++)                                     //  no pixels mapped yet
      for (dx = 0; dx <= 2*rad1; dx++)
         pixseq_done[dx][dy] = 0;

      ii = jj = 0;
      
      astep = 0.5 / rad1;                                                  //  0.5 pixel steps at rad1 from center

      for (angle = 0; angle < 2*pi; angle += astep)                        //  loop full circle
      {
         pixseq_angle[ii] = jj;                                            //  start pixel sequence for this angle
         ii++;

         for (rad2 = 1; rad2 <= rad1; rad2++)                              //  loop rad2 from center to edge
         {
            dx = lround(rad2 * cos(angle));                                //  pixel at angle and rad2
            dy = lround(rad2 * sin(angle));
            adx = rad1 + dx;
            ady = rad1 + dy;
            if (pixseq_done[adx][ady]) continue;                           //  pixel already mapped
            pixseq_done[adx][ady] = 1;                                     //  map pixel
            pixseq_dx[jj] = dx;                                            //  save pixel sequence for angle
            pixseq_dy[jj] = dy;
            pixseq_rad[jj] = rad2;                                         //  pixel radius
            jj++;
         }
         pixseq_rad[jj] = 9999;                                            //  mark end of pixels for angle
         jj++;
      }
      
      pixseq_angle[ii] = 9999;                                             //  mark end of angle steps

      if (ii > max1 || jj > max2)                                          //  should not happen
         zappcrash("gradblur array overflow");

      if (sa_stat == 3) Fbusy_goal = sa_Npixel;                            //  setup progress tracking
      else  Fbusy_goal = ww * hh;
      Fbusy_done = 0;

      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(gradblur_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      Fbusy_goal = 0;
      CEF->Fmods++; 
      CEF->Fsaved = 0;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * gradblur_wthread(void *arg)                                         //  worker thread function
{
   using namespace gradblur_names;

   int      index = *((int *) arg);
   int      ii, jj, npix, dist = 0;
   int      px, py, dx, dy;
   int      ww = E3pxm->ww, hh = E3pxm->hh;
   float    red, green, blue, f1, f2;
   float    *pix1, *pix3, *pixN;
   int      rad, blurrad, con;
   
   for (py = index+1; py < hh-1; py += Nwt)                                //  loop interior pixels
   for (px = 1; px < ww-1; px++)
   {
      if (gradblur_cancel) exit_wthread();
      
      ii = py * ww + px;
         
      if (sa_stat == 3) {                                                  //  select area active
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel outside area
      }

      if (pixcon[ii] > con_limit) continue;                                //  high contrast pixel

      pix1 = PXMpix(E1pxm,px,py);                                          //  source pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  target pixel

      red = pix1[0];                                                       //  blur center pixel
      green = pix1[1];
      blue = pix1[2];
      npix = 1;

      blurrad = 1.0 + blur_radius * (con_limit - pixcon[ii]) / con_limit;  //  blur radius for pixel, 1 ... blur_radius

      for (ii = 0; ii < 2000; ii++)                                        //  loop angle around center pixel
      {
         jj = pixseq_angle[ii];                                            //  1st pixel for angle step ii
         if (jj == 9999) break;                                            //  none, end of angle loop
         
         while (true)                                                      //  loop pixels from center to radius
         {
            rad = pixseq_rad[jj];                                          //  next pixel step radius
            if (rad > blurrad) break;                                      //  stop here if beyond blur radius
            dx = pixseq_dx[jj];                                            //  next pixel step px/py
            dy = pixseq_dy[jj];
            if (px+dx < 0 || px+dx > ww-1) break;                          //  stop here if off the edge
            if (py+dy < 0 || py+dy > hh-1) break;
            
            pixN = PXMpix(E1pxm,px+dx,py+dy);
            con = 255 * pixcontrast(pix1,pixN);
            if (con > con_limit) break;

            red += pixN[0];                                                //  add pixel RGB values
            green += pixN[1];
            blue += pixN[2];
            npix++;
            jj++;
         }
      }

      pix3[0] = red / npix;                                                //  average pixel values
      pix3[1] = green / npix;
      pix3[2] = blue / npix;

      if (sa_stat == 3 && dist < sa_blend) {
         f1 = 1.0 * dist / sa_blend;
         f2 = 1.0 - f1;
         pix3[0] = f1 * pix3[0] + f2 * pix1[0];
         pix3[1] = f1 * pix3[1] + f2 * pix1[1];
         pix3[2] = f1 * pix3[2] + f2 * pix1[2];
      }

      Fbusy_done++;                                                        //  track progress
   }
   
   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  convert image to simulate an embossing

int      emboss_radius, emboss_color;
float    emboss_depth;
float    emboss_kernel[20][20];                                            //  support radius <= 9

editfunc    EFemboss;


void m_emboss(GtkWidget *, cchar *)
{
   int    emboss_dialog_event(zdialog* zd, cchar *event);
   void * emboss_thread(void *);

   cchar  *title = ZTX("Simulate Embossing");

   F1_help_topic = "embossing";

   EFemboss.funcname = "emboss";
   EFemboss.Farea = 2;                                                     //  select area usable
   EFemboss.threadfunc = emboss_thread;                                    //  thread function
   if (! edit_setup(EFemboss)) return;                                     //  setup edit

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);
   EFemboss.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zd,"label","lab1","hb1",Bradius,"space=5");
   zdialog_add_widget(zd,"spin","radius","hb1","0|9|1|0");
   zdialog_add_widget(zd,"label","lab2","hb1",ZTX("depth"),"space=5");
   zdialog_add_widget(zd,"spin","depth","hb1","0|99|1|0");
   zdialog_add_widget(zd,"check","color","hb1",ZTX("color"),"space=8");

   zdialog_run(zd,emboss_dialog_event,"-10/20");                           //  run dialog, parallel
   return;
}


//  dialog event and completion callback function

int emboss_dialog_event(zdialog *zd, cchar *event)                         //  emboss dialog event function
{
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 0;
   }

   if (strEqu(event,"undo")) edit_undo();
   if (strEqu(event,"redo")) edit_redo();
   if (strEqu(event,"blendwidth")) signal_thread();

   if (strcmpv(event,"radius","depth","color",null))
   {
      zdialog_fetch(zd,"radius",emboss_radius);                            //  get user inputs
      zdialog_fetch(zd,"depth",emboss_depth);
      zdialog_fetch(zd,"color",emboss_color);
      signal_thread();                                                     //  trigger update thread
   }

   return 1;
}


//  thread function - use multiple working threads

void * emboss_thread(void *)
{
   void  * emboss_wthread(void *arg);

   int         ii, dx, dy, rad;
   float       depth, kern, coeff;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      rad = emboss_radius;
      depth = emboss_depth;

      coeff = 0.1 * depth / (rad * rad + 1);

      for (dy = -rad; dy <= rad; dy++)                                     //  build kernel with radius and depth
      for (dx = -rad; dx <= rad; dx++)
      {
         kern = coeff * (dx + dy);
         emboss_kernel[dx+rad][dy+rad] = kern;
      }
      
      emboss_kernel[rad][rad] = 1;                                         //  kernel center cell = 1

      for (ii = 0; ii < Nwt; ii++)                                         //  start worker threads
         start_wthread(emboss_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for comletion

      CEF->Fmods++; 
      CEF->Fsaved = 0;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * emboss_wthread(void *arg)                                           //  worker thread function
{
   void  emboss_1pix(int px, int py);

   int         index = *((int *) (arg));
   int         px, py;

   for (py = index; py < E1pxm->hh; py += Nwt)                             //  process all pixels
   for (px = 0; px < E1pxm->ww; px++)
      emboss_1pix(px,py);

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


void emboss_1pix(int px, int py)                                           //  process one pixel
{
   int         ii, dist = 0;
   float       bright1, bright3;
   int         rgb, dx, dy, rad;
   float       *pix1, *pix3, *pixN;
   float       sumpix, kern, dold, dnew;
   
   if (sa_stat == 3) {                                                     //  select area active
      ii = py * Fpxb->ww + px;
      dist = sa_pixmap[ii];                                                //  distance from edge
      if (! dist) return;                                                  //  outside pixel
   }

   rad = emboss_radius;

   if (px < rad || py < rad) return;
   if (px > E3pxm->ww-rad-1 || py > E3pxm->hh-rad-1) return;
   
   pix1 = PXMpix(E1pxm,px,py);                                             //  input pixel
   pix3 = PXMpix(E3pxm,px,py);                                             //  output pixel
   
   if (emboss_color) 
   {
      for (rgb = 0; rgb < 3; rgb++)
      {      
         sumpix = 0;
         
         for (dy = -rad; dy <= rad; dy++)                                  //  loop surrounding block of pixels
         for (dx = -rad; dx <= rad; dx++)
         {
            pixN = pix1 + (dy * E1pxm->ww + dx) * 3;
            kern = emboss_kernel[dx+rad][dy+rad];
            sumpix += kern * pixN[rgb];
      
            bright1 = pix1[rgb];
            bright3 = sumpix;
            if (bright3 < 0) bright3 = 0;
            if (bright3 > 255) bright3 = 255;

            if (sa_stat == 3 && dist < sa_blend) {                         //  select area is active,
               dnew = 1.0 * dist / sa_blend;                               //    blend changes over sa_blend
               dold = 1.0 - dnew;
               bright3 = dnew * bright3 + dold * bright1;
            }

            pix3[rgb] = bright3;
         }
      }
   }
   
   else                                                                    //  use gray scale
   {
      sumpix = 0;
         
      for (dy = -rad; dy <= rad; dy++)                                     //  loop surrounding block of pixels
      for (dx = -rad; dx <= rad; dx++)
      {
         pixN = pix1 + (dy * E1pxm->ww + dx) * 3;
         kern = emboss_kernel[dx+rad][dy+rad];
         sumpix += kern * (pixN[0] + pixN[1] + pixN[2]);
      }
      
      bright1 = 0.3333 * (pix1[0] + pix1[1] + pix1[2]);
      bright3 = 0.3333 * sumpix;
      if (bright3 < 0) bright3 = 0;
      if (bright3 > 255) bright3 = 255;
      
      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         dnew = 1.0 * dist / sa_blend;                                     //    blend changes over sa_blend
         dold = 1.0 - dnew;
         bright3 = dnew * bright3 + dold * bright1;
      }

      pix3[0] = pix3[1] = pix3[2] = bright3;
   }

   return;
}


/**************************************************************************/

//  convert image to simulate square tiles

int         tile_size, tile_gap, tile_3D;
float       *tile_pixmap = 0;

editfunc    EFtiles;


void m_tiles(GtkWidget *, cchar *)
{
   int    tiles_dialog_event(zdialog *zd, cchar *event);
   void * tiles_thread(void *);

   F1_help_topic = "tiles";

   EFtiles.funcname = "tiles";
   EFtiles.Farea = 2;                                                      //  select area usable
   EFtiles.threadfunc = tiles_thread;                                      //  thread function
   if (! edit_setup(EFtiles)) return;                                      //  setup edit

   zdialog *zd = zdialog_new(ZTX("Simulate Tiles"),mWin,Bapply,Bdone,Bcancel,null);
   EFtiles.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labt","hb1",ZTX("tile size"),"space=5");
   zdialog_add_widget(zd,"spin","size","hb1","1|99|1|10","space=5");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labg","hb2",ZTX("tile gap"),"space=5");
   zdialog_add_widget(zd,"spin","gap","hb2","0|9|1|1","space=5");
   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labd","hb3",ZTX("3D depth"),"space=5");
   zdialog_add_widget(zd,"spin","3D","hb3","0|9|1|1","space=5");

   zdialog_run(zd,tiles_dialog_event,"-10/20");                            //  run dialog, parallel

   tile_size = 10;                                                         //  defaults
   tile_gap = 1;
   tile_3D = 0;

   int cc = E1pxm->ww * E1pxm->hh * 3 * sizeof(float);
   tile_pixmap = (float *) zmalloc(cc,"tiles");                            //  set up pixel color map
   memset(tile_pixmap,0,cc);

   return;
}


//  dialog event and completion callback function

int tiles_dialog_event(zdialog * zd, cchar *event)
{
   if (zd->zstat && zd->zstat != 1)                                        //  dialog complete
   {
      if (zd->zstat == 2) edit_done();                                     //  done
      else edit_cancel();                                                  //  cancel
      if (zd->zstat) {
         zfree(tile_pixmap);
         return 0;
      }
   }

   if (strEqu(event,"blendwidth")) signal_thread();

   if (zd->zstat != 1) return 0;                                           //  apply
   zd->zstat = 0;                                                          //  keep dialog active

   zdialog_fetch(zd,"size",tile_size);                                     //  get tile size
   zdialog_fetch(zd,"gap",tile_gap);                                       //  get tile gap
   zdialog_fetch(zd,"3D",tile_3D);                                         //  get 3D yes/no

   if (tile_size < 2) {
      edit_reset();                                                        //  restore original image
      return 0;
   }
   
   signal_thread();                                                        //  trigger working thread
   return 1;
}


//  image tiles thread function

void * tiles_thread(void *)
{
   int         sg, gg, eg, sumpix;
   int         ii, jj, px, py, qx, qy, dist;
   int         td, bd, ld, rd;
   float       red = 0, green = 0, blue = 0;
   float       hired = 0, higreen = 0, hiblue = 0;
   float       lored = 0, logreen = 0, loblue = 0;
   float       dnew, dold;
   float       *pix1, *pix3;

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      sg = tile_size + tile_gap;
      gg = tile_gap;

      for (py = 0; py < E1pxm->hh; py += sg)                               //  initz. pixel color map for
      for (px = 0; px < E1pxm->ww; px += sg)                               //    given pixel size
      {
         sumpix = red = green = blue = 0;

         for (qy = py + gg; qy < py + sg; qy++)                            //  get mean color for pixel block
         for (qx = px + gg; qx < px + sg; qx++)
         {
            if (qy > E1pxm->hh-1 || qx > E1pxm->ww-1) continue;

            pix1 = PXMpix(E1pxm,qx,qy);
            red += pix1[0];
            green += pix1[1];
            blue += pix1[2];
            sumpix++;
         }

         if (! sumpix) continue;

         red = (red / sumpix);                                             //  tile color
         green = (green / sumpix);
         blue = (blue / sumpix);
         
         if (tile_3D)
         {
            hired = 0.5 * (red + 255.0);                                   //  3D highlight and shadow colors
            higreen = 0.5 * (green + 255.0);
            hiblue = 0.5 * (blue + 255.0);
            lored = 0.5 * red;
            logreen = 0.5 * green;
            loblue = 0.5 * blue;
         }
         
         for (qy = py; qy < py + sg; qy++)                                 //  set color for pixels in block
         for (qx = px; qx < px + sg; qx++)
         {
            if (qy > E1pxm->hh-1 || qx > E1pxm->ww-1) continue;
            
            jj = (qy * E1pxm->ww + qx) * 3;                                //  pixel index
            
            td = qx - px;                                                  //  distance from top edge of tile
            bd = px + sg - qx;                                             //  distance from bottom edge
            ld = qy - py;                                                  //  distance from left edge
            rd = py + sg - qy;                                             //  distance from right edge

            if (td < gg || ld < gg) {                                      //  gap pixels are black
               tile_pixmap[jj] = tile_pixmap[jj+1] = tile_pixmap[jj+2] = 0;
               continue;
            }
            
            if (tile_3D)
            {
               eg = tile_3D;                                               //  tile "depth"
               
               if (td < gg+eg && td < rd) {                                //  top edge: highlight
                  tile_pixmap[jj] = hired;                                 //  bevel at right end
                  tile_pixmap[jj+1] = higreen;
                  tile_pixmap[jj+2] = hiblue;
                  continue;
               }
               
               if (ld < gg+eg && ld < bd) {                                //  left edge: highlight
                  tile_pixmap[jj] = hired;                                 //  bevel at bottom end
                  tile_pixmap[jj+1] = higreen;
                  tile_pixmap[jj+2] = hiblue;
                  continue;
               }
               
               if (bd <= eg || rd <= eg) {                                 //  bottom or right edge: shadow
                  tile_pixmap[jj] = lored;
                  tile_pixmap[jj+1] = logreen;
                  tile_pixmap[jj+2] = loblue;
                  continue;
               }
            }

            tile_pixmap[jj] = red;
            tile_pixmap[jj+1] = green;
            tile_pixmap[jj+2] = blue;
         }
      }

      if (sa_stat == 3)                                                    //  process selected area
      {
         for (ii = 0; ii < Fpxb->ww * Fpxb->hh; ii++)                      //  find pixels in select area
         {
            if (! sa_pixmap[ii]) continue;
            py = ii / Fpxb->ww;
            px = ii - py * Fpxb->ww;
            dist = sa_pixmap[ii];                                          //  distance from edge
            pix3 = PXMpix(E3pxm,px,py);
            jj = (py * E3pxm->ww + px) * 3;

            if (dist >= sa_blend) {                                        //  blend changes over sa_blend 
               pix3[0] = tile_pixmap[jj];
               pix3[1] = tile_pixmap[jj+1];                                //  apply block color to member pixels
               pix3[2] = tile_pixmap[jj+2];
            }
            else {
               dnew = 1.0 * dist / sa_blend;
               dold = 1.0 - dnew;
               pix1 = PXMpix(E1pxm,px,py);
               pix3[0] = dnew * tile_pixmap[jj] + dold * pix1[0];
               pix3[1] = dnew * tile_pixmap[jj+1] + dold * pix1[1];
               pix3[2] = dnew * tile_pixmap[jj+2] + dold * pix1[2];
            }
         }
      }

      else                                                                 //  process entire image
      {
         for (py = 0; py < E3pxm->hh-1; py++)                              //  loop all image pixels
         for (px = 0; px < E3pxm->ww-1; px++)
         {
            pix3 = PXMpix(E3pxm,px,py);                                    //  target pixel
            jj = (py * E3pxm->ww + px) * 3;                                //  color map for (px,py)
            pix3[0] = tile_pixmap[jj];
            pix3[1] = tile_pixmap[jj+1];
            pix3[2] = tile_pixmap[jj+2];
         }
      }

      CEF->Fmods++; 
      CEF->Fsaved = 0;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************/

//  convert image to dot grid (like Roy Lichtenstein)

int      dot_size;                                                         //  tile size enclosing dots

editfunc    EFdots;


void m_dots(GtkWidget *, cchar *)
{
   int    dots_dialog_event(zdialog *zd, cchar *event);
   void * dots_thread(void *);

   F1_help_topic = "dot_matrix";

   EFdots.funcname = "dots";
   EFdots.Farea = 2;                                                       //  select area usable
   EFdots.threadfunc = dots_thread;                                        //  thread function
   if (! edit_setup(EFdots)) return;                                       //  setup edit

   zdialog *zd = zdialog_new(ZTX("Convert Image to Dots"),mWin,Bdone,Bcancel,null);
   EFdots.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labt","hb1",ZTX("dot size"),"space=5");
   zdialog_add_widget(zd,"spin","size","hb1","4|20|1|9","space=5");
   zdialog_add_widget(zd,"button","apply","hb1",Bapply,"space=10");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");

   zdialog_run(zd,dots_dialog_event,"-10/20");                             //  run dialog, parallel

   dot_size = 9;
   return;
}


//  dialog event and completion callback function

int dots_dialog_event(zdialog * zd, cchar *event)
{
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 0;
   }

   if (strEqu(event,"undo")) edit_undo();
   if (strEqu(event,"redo")) edit_redo();
   if (strEqu(event,"blendwidth")) signal_thread();
   if (strNeq(event,"apply")) return 0;                                    //  wait for [apply]

   zdialog_fetch(zd,"size",dot_size);                                      //  get dot size
   signal_thread();                                                        //  trigger working thread
   return 1;
}


//  image dots thread function

void * dots_thread(void *)
{
   int         ds, cc, ii, sumpix;
   int         dist, shift1, shift2;
   int         px, py, px1, px2, py1, py2;
   int         qx, qy, qx1, qx2, qy1, qy2;
   float       *pix1, *pix3;
   float       red, green, blue, maxrgb;
   float       relmax, radius, radius2a, radius2b, f1, f2;
   float       fpi = 1.0 / 3.14159;
   float       dcx, dcy, qcx, qcy, qdist2;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      mutex_lock(&Fpixmap_lock); 

      ds = dot_size;                                                       //  dot size and tile size

      cc = E3pxm->ww * E3pxm->hh * 3 * sizeof(float);                      //  clear output image to black
      memset(E3pxm->pixels,0,cc);
      
      px1 = 0;                                                             //  limits for tiles 
      px2 = Fpxb->ww - ds;
      py1 = 0;
      py2 = Fpxb->hh - ds;
      
      if (sa_stat == 3) {                                                  //  reduce for active area
         px1 = sa_minx;
         px2 = sa_maxx;
         py1 = sa_miny;
         py2 = sa_maxy;
         if (px2 > Fpxb->ww - ds) px2 = Fpxb->ww - ds;
         if (py2 > Fpxb->hh - ds) py2 = Fpxb->hh - ds;
      }
      
      shift1 = 0;

      for (py = py1; py < py2; py += ds)                                   //  loop all tiles in input image
      {
         shift1 = 1 - shift1;                                              //  offset alternate rows
         shift2 = 0.5 * shift1 * ds;

         for (px = px1 + shift2; px < px2; px += ds)
         {
            sumpix = red = green = blue = 0;

            for (qy = py; qy < py + ds; qy++)                              //  loop all pixels in tile
            for (qx = px; qx < px + ds; qx++)                              //  get mean RGB levels for tile
            {
               pix1 = PXMpix(E1pxm,qx,qy);
               red += pix1[0];
               green += pix1[1];
               blue += pix1[2];
               sumpix++;
            }
            
            red = red / sumpix;                                            //  mean RGB levels, 0 to 255.9
            green = green / sumpix;
            blue = blue / sumpix;
            
            maxrgb = red;                                                  //  max. mean RGB level, 0 to 255.9
            if (green > maxrgb) maxrgb = green;
            if (blue > maxrgb) maxrgb = blue;
            relmax = maxrgb / 256.0;                                       //  max. RGB as 0 to 0.999
            
            radius = ds * sqrt(fpi * relmax);                              //  radius of dot with maximized color
            radius = radius * 0.875;                                       //  deliberate reduction         v.13.08
            if (radius < 0.5) continue;
            
            red = 255.9 * red / maxrgb;                                    //  dot color, maximized
            green = 255.9 * green / maxrgb;
            blue = 255.9 * blue / maxrgb;

            dcx = px + 0.5 * ds;                                           //  center of dot / tile
            dcy = py + 0.5 * ds;
            
            qx1 = dcx - radius;                                            //  pixels within dot radius of center
            qx2 = dcx + radius;
            qy1 = dcy - radius;
            qy2 = dcy + radius;
            
            radius2a = (radius + 0.5) * (radius + 0.5);
            radius2b = (radius - 0.5) * (radius - 0.5);
            
            for (qy = qy1; qy <= qy2; qy++)                                //  loop all pixels within dot radius
            for (qx = qx1; qx <= qx2; qx++)
            {
               qcx = qx + 0.5;                                             //  center of pixel
               qcy = qy + 0.5;
               qdist2 = (qcx-dcx)*(qcx-dcx) + (qcy-dcy)*(qcy-dcy);         //  pixel distance**2 from center of dot
               if (qdist2 > radius2a) f1 = 0.0;                            //  pixel outside dot, no color
               else if (qdist2 < radius2b) f1 = 1.0;                       //  pixel inside dot, full color
               else f1 = 1.0 - sqrt(qdist2) + radius - 0.5;                //  pixel straddles edge, some color
               pix3 = PXMpix(E3pxm,qx,qy);
               pix3[0] = f1 * red;
               pix3[1] = f1 * green;
               pix3[2] = f1 * blue;
            }
         }
      }
      
      if (sa_stat == 3)                                                    //  select area active
      {
         pix1 = E1pxm->pixels;
         pix3 = E3pxm->pixels;
         
         for (ii = 0; ii < Fpxb->ww * Fpxb->hh; ii++)                      //  loop all pixels
         {
            dist = sa_pixmap[ii];
            if (dist == 0) memmove(pix3,pix1,3*sizeof(float));             //  pixel outside area, output pixel = input
            else if (dist < sa_blend) {
               f1 = 1.0 * dist / sa_blend;                                 //  pixel in blend region
               f2 = 1.0 - f1;
               pix3[0] = f1 * pix3[0] + f2 * pix1[0];                      //  output pixel is mix of input and output
               pix3[1] = f1 * pix3[1] + f2 * pix1[1];
               pix3[2] = f1 * pix3[2] + f2 * pix1[2];
            }
            pix1 += 3;
            pix3 += 3;
         }
      }

      mutex_unlock(&Fpixmap_lock); 

      CEF->Fmods++; 
      CEF->Fsaved = 0;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************/

//  convert image to simulate a painting
//  processing a 10 megapixel image needs 140 MB of main memory

namespace painting_names 
{
   editfunc    EFpainting;

   int         color_depth;
   int         group_area;
   float       color_match;
   int         borders;

   typedef struct  {
      int16       px, py;
      char        direc;
   }  spixstack;

   int         Nstack;
   spixstack   *pixstack;                                                  //  pixel group search memory
   int         *pixgroup;                                                  //  maps (px,py) to pixel group no.
   int         *groupcount;                                                //  count of pixels in each group

   int         group;
   char        direc;
   float       gcolor[3];
}


void m_painting(GtkWidget *, cchar *)
{
   using namespace painting_names;

   int    painting_dialog_event(zdialog *zd, cchar *event);
   void * painting_thread(void *);

   F1_help_topic = "painting";

   EFpainting.funcname = "painting";
   EFpainting.Farea = 2;                                                   //  select area usable
   EFpainting.threadfunc = painting_thread;                                //  thread function
   if (! edit_setup(EFpainting)) return;                                   //  setup edit

   zdialog *zd = zdialog_new(ZTX("Simulate Painting"),mWin,Bdone,Bcancel,null);
   EFpainting.zd = zd;

   zdialog_add_widget(zd,"hbox","hbcd","dialog",0,"space=1");
   zdialog_add_widget(zd,"label","lab1","hbcd",ZTX("color depth"),"space=5");
   zdialog_add_widget(zd,"spin","colordepth","hbcd","1|5|1|3","space=5");

   zdialog_add_widget(zd,"hbox","hbts","dialog",0,"space=1");
   zdialog_add_widget(zd,"label","labts","hbts",ZTX("patch area goal"),"space=5");
   zdialog_add_widget(zd,"spin","grouparea","hbts","0|9999|10|1000","space=5");

   zdialog_add_widget(zd,"hbox","hbcm","dialog",0,"space=1");
   zdialog_add_widget(zd,"label","labcm","hbcm",ZTX("req. color match"),"space=5");
   zdialog_add_widget(zd,"spin","colormatch","hbcm","0|99|1|50","space=5");

   zdialog_add_widget(zd,"hbox","hbbd","dialog",0,"space=1");
   zdialog_add_widget(zd,"label","labbd","hbbd",ZTX("borders"),"space=5");
   zdialog_add_widget(zd,"check","borders","hbbd",0,"space=2");
   zdialog_add_widget(zd,"button","apply","hbbd",Bapply,"space=10");

   zdialog_restore_inputs(zd);                                             //  preload prior user inputs
   zdialog_run(zd,painting_dialog_event,"save");                           //  run dialog, parallel

   return;
}


//  dialog event and completion callback function

int painting_dialog_event(zdialog *zd, cchar *event)
{
   using namespace painting_names;

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 0;
   }

   if (strEqu(event,"undo")) edit_undo();
   if (strEqu(event,"redo")) edit_redo();
   if (strEqu(event,"blendwidth")) signal_thread();

   if (strEqu(event,"apply")) {                                            //  apply user settings
      zdialog_fetch(zd,"colordepth",color_depth);                          //  color depth
      zdialog_fetch(zd,"grouparea",group_area);                            //  target group area (pixels)
      zdialog_fetch(zd,"colormatch",color_match);                          //  req. color match to combine groups
      zdialog_fetch(zd,"borders",borders);                                 //  borders wanted
      color_match = 0.01 * color_match;                                    //  scale 0 to 1

      signal_thread();                                                     //  do the work
      wait_thread_idle();                                                  //  v.13.09
   }
   
   return 0;
}


//  painting thread function

void * painting_thread(void *)
{
   void  painting_colordepth();
   void  painting_pixgroups();
   void  painting_mergegroups();
   void  painting_paintborders();
   void  painting_blend();

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      Ffuncbusy++;

      painting_colordepth();                                               //  set new color depth
      painting_pixgroups();                                                //  group pixel patches of a color
      painting_mergegroups();                                              //  merge smaller into larger groups
      painting_paintborders();                                             //  add borders around groups
      painting_blend();                                                    //  blend edges of selected area

      Ffuncbusy--;
      CEF->Fmods++; 
      CEF->Fsaved = 0;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


//  set the specified color depth, 1-5 bits/color

void painting_colordepth()
{
   using namespace painting_names;

   int            ii, px, py, rgb;
   uint16         m1, m2, val1, val3;
   float          fmag, *pix1, *pix3;

   m1 = 0xFFFF << (16 - color_depth);                                      //  5 > 1111100000000000
   m2 = 0x8000 >> color_depth;                                             //  5 > 0000010000000000

   fmag = 65535.0 / m1;                                                    //  full brightness range

   for (py = 0; py < E3pxm->hh; py++)                                      //  loop all pixels
   for (px = 0; px < E3pxm->ww; px++)
   {
      if (sa_stat == 3) {                                                  //  select area active
         ii = py * Fpxb->ww + px;
         if (! sa_pixmap[ii]) continue;                                    //  outside pixel
      }
      
      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel
      
      for (rgb = 0; rgb < 3; rgb++)
      {
         val1 = 256 * pix1[rgb];
         if (val1 < m1) val3 = (val1 + m2) & m1;
         else val3 = m1;
         val3 = val3 * fmag;
         pix3[rgb] = val3 / 256;
      }
   }

   return;
}


//  find all groups of contiguous pixels with the same color

void painting_pixgroups()
{
   using namespace painting_names;

   void  painting_pushpix(int px, int py);

   int            cc1, cc2;
   int            ii, kk, px, py;
   int            ppx, ppy, npx, npy;
   float          *pix3;

   cc1 = E3pxm->ww * E3pxm->hh;

   cc2 = cc1 * sizeof(int);
   pixgroup = (int *) zmalloc(cc2,"painting");                             //  maps pixel to assigned group
   memset(pixgroup,0,cc2);
   
   if (sa_stat == 3) cc1 = sa_Npixel;

   cc2 = cc1 * sizeof(spixstack);
   pixstack = (spixstack *) zmalloc(cc2,"painting");                       //  memory stack for pixel search
   memset(pixstack,0,cc2);
   
   cc2 = cc1 * sizeof(int);
   groupcount = (int *) zmalloc(cc2,"painting");                           //  counts pixels per group
   memset(groupcount,0,cc2);
   
   group = 0;
   
   for (py = 0; py < E3pxm->hh; py++)                                      //  loop all pixels
   for (px = 0; px < E3pxm->ww; px++)
   {
      kk = py * E3pxm->ww + px;
      if (sa_stat == 3 && ! sa_pixmap[kk]) continue;
      if (pixgroup[kk]) continue;                                          //  already assigned to group

      pixgroup[kk] = ++group;                                              //  assign next group
      ++groupcount[group];

      pix3 = PXMpix(E3pxm,px,py);
      gcolor[0] = pix3[0];
      gcolor[1] = pix3[1];
      gcolor[2] = pix3[2];

      pixstack[0].px = px;                                                 //  put pixel into stack with
      pixstack[0].py = py;                                                 //    direction = ahead
      pixstack[0].direc = 'a';
      Nstack = 1;

      while (Nstack)                                                       //  overhauled
      {
         kk = Nstack - 1;                                                  //  get last pixel in stack
         px = pixstack[kk].px;
         py = pixstack[kk].py;
         direc = pixstack[kk].direc;
         
         if (direc == 'x') {
            Nstack--;
            continue;
         }

         if (Nstack > 1) {
            ii = Nstack - 2;                                               //  get prior pixel in stack
            ppx = pixstack[ii].px;
            ppy = pixstack[ii].py;
         }
         else {
            ppx = px - 1;                                                  //  if only one, assume prior = left
            ppy = py;
         }

         if (direc == 'a') {                                               //  next ahead pixel
            npx = px + px - ppx;
            npy = py + py - ppy;
            pixstack[kk].direc = 'r';                                      //  next search direction
         }

         else if (direc == 'r') {                                          //  next right pixel 
            npx = px + py - ppy;
            npy = py + px - ppx;
            pixstack[kk].direc = 'l';
         }

         else { /*  direc = 'l'  */                                        //  next left pixel
            npx = px + ppy - py;
            npy = py + ppx - px;
            pixstack[kk].direc = 'x';
         }

         if (npx < 0 || npx > E3pxm->ww-1) continue;                       //  pixel off the edge
         if (npy < 0 || npy > E3pxm->hh-1) continue;
         
         kk = npy * E3pxm->ww + npx;

         if (sa_stat == 3 && ! sa_pixmap[kk]) continue;                    //  pixel outside area

         if (pixgroup[kk]) continue;                                       //  pixel already assigned

         pix3 = PXMpix(E3pxm,npx,npy);
         if (pix3[0] != gcolor[0] || pix3[1] != gcolor[1]                  //  not same color as group
                                  || pix3[2] != gcolor[2]) continue;
         
         pixgroup[kk] = group;                                             //  assign pixel to group
         ++groupcount[group];

         kk = Nstack++;                                                    //  put pixel into stack
         pixstack[kk].px = npx;
         pixstack[kk].py = npy;
         pixstack[kk].direc = 'a';                                         //  direction = ahead
      }
   }
   
   return;
}      


//  merge small pixel groups into adjacent larger groups with best color match

void painting_mergegroups()
{
   using namespace painting_names;

   int         ii, jj, kk, px, py, npx, npy;
   int         nccc, mcount, group2;
   int         nnpx[4] = {  0, -1, +1, 0 };
   int         nnpy[4] = { -1, 0,  0, +1 };
   float       ff = 1.0 / 256.0;
   float       fred, fgreen, fblue, match;
   float       *pix3, *pixN;

   typedef struct  {
      int         group;
      double      match;
      float       pixM[3];
   }  snewgroup;

   snewgroup      *newgroup;
   
   nccc = (group + 1) * sizeof(snewgroup);
   newgroup = (snewgroup *) zmalloc(nccc,"painting");
   
   if (sa_stat == 3)                                                       //  process select area
   {
      while (true)
      {
         memset(newgroup,0,nccc);

         for (ii = 0; ii < Fpxb->ww * Fpxb->hh; ii++)                      //  find pixels in select area
         {
            if (! sa_pixmap[ii]) continue;
            py = ii / Fpxb->ww;
            px = ii - py * Fpxb->ww;

            kk = E3pxm->ww * py + px;                                      //  get assigned group
            group = pixgroup[kk];
            if (groupcount[group] >= group_area) continue;                 //  group count large enough

            pix3 = PXMpix(E3pxm,px,py);

            for (jj = 0; jj < 4; jj++)                                     //  get 4 neighbor pixels
            {
               npx = px + nnpx[jj];
               npy = py + nnpy[jj];

               if (npx < 0 || npx >= E3pxm->ww) continue;                  //  off the edge
               if (npy < 0 || npy >= E3pxm->hh) continue;

               kk = E3pxm->ww * npy + npx;
               if (! sa_pixmap[kk]) continue;                              //  pixel outside area
               if (pixgroup[kk] == group) continue;                        //  already in same group
               
               pixN = PXMpix(E3pxm,npx,npy);                               //  match color of group neighbor
               fred = ff * fabsf(pix3[0] - pixN[0]);                       //    to color of group
               fgreen = ff * fabsf(pix3[1] - pixN[1]);
               fblue = ff * fabsf(pix3[2] - pixN[2]);
               match = (1.0 - fred) * (1.0 - fgreen) * (1.0 - fblue);      //  color match, 0 to 1.0
               if (match < color_match) continue;

               if (match > newgroup[group].match) {
                  newgroup[group].match = match;                           //  remember best match
                  newgroup[group].group = pixgroup[kk];                    //  and corresp. group no.
                  newgroup[group].pixM[0] = pixN[0];                       //  and corresp. new color
                  newgroup[group].pixM[1] = pixN[1];
                  newgroup[group].pixM[2] = pixN[2];
               }
            }
         }

         mcount = 0;

         for (ii = 0; ii < Fpxb->ww * Fpxb->hh; ii++)                      //  find pixels in select area
         {
            if (! sa_pixmap[ii]) continue;
            py = ii / Fpxb->ww;
            px = ii - py * Fpxb->ww;

            kk = E3pxm->ww * py + px;
            group = pixgroup[kk];                                          //  test for new group assignment
            group2 = newgroup[group].group;
            if (! group2) continue;
            
            if (groupcount[group] > groupcount[group2]) continue;          //  accept only bigger new group

            pixgroup[kk] = group2;                                         //  make new group assignment
            --groupcount[group];
            ++groupcount[group2];

            pix3 = PXMpix(E3pxm,px,py);                                    //  make new color assignment
            pix3[0] = newgroup[group].pixM[0];
            pix3[1] = newgroup[group].pixM[1];
            pix3[2] = newgroup[group].pixM[2];

            mcount++;
         }
         
         if (mcount == 0) break;
      }
   }

   else                                                                    //  process entire image
   {
      while (true)
      {
         memset(newgroup,0,nccc);

         for (py = 0; py < E3pxm->hh; py++)                                //  loop all pixels
         for (px = 0; px < E3pxm->ww; px++)
         {
            kk = E3pxm->ww * py + px;                                      //  get assigned group
            group = pixgroup[kk];
            if (groupcount[group] >= group_area) continue;                 //  group count large enough

            pix3 = PXMpix(E3pxm,px,py);

            for (jj = 0; jj < 4; jj++)                                     //  get 4 neighbor pixels
            {
               npx = px + nnpx[jj];
               npy = py + nnpy[jj];

               if (npx < 0 || npx >= E3pxm->ww) continue;                  //  off the edge
               if (npy < 0 || npy >= E3pxm->hh) continue;
               
               kk = E3pxm->ww * npy + npx;
               if (pixgroup[kk] == group) continue;                        //  in same group

               pixN = PXMpix(E3pxm,npx,npy);                               //  match color of group neighbor
               fred = ff * fabsf(pix3[0] - pixN[0]);                       //    to color of group
               fgreen = ff * fabsf(pix3[1] - pixN[1]);
               fblue = ff * fabsf(pix3[2] - pixN[2]);
               match = (1.0 - fred) * (1.0 - fgreen) * (1.0 - fblue);      //  color match, 0 to 1.0
               if (match < color_match) continue;

               if (match > newgroup[group].match) {
                  newgroup[group].match = match;                           //  remember best match
                  newgroup[group].group = pixgroup[kk];                    //  and corresp. group no.
                  newgroup[group].pixM[0] = pixN[0];                       //  and corresp. new color
                  newgroup[group].pixM[1] = pixN[1];
                  newgroup[group].pixM[2] = pixN[2];
               }
            }
         }

         mcount = 0;

         for (py = 0; py < E3pxm->hh; py++)                                //  loop all pixels
         for (px = 0; px < E3pxm->ww; px++)
         {
            kk = E3pxm->ww * py + px;
            group = pixgroup[kk];                                          //  test for new group assignment
            group2 = newgroup[group].group;
            if (! group2) continue;
            
            if (groupcount[group] > groupcount[group2]) continue;          //  accept only bigger new group

            pixgroup[kk] = group2;                                         //  make new group assignment
            --groupcount[group];
            ++groupcount[group2];

            pix3 = PXMpix(E3pxm,px,py);                                    //  make new color assignment
            pix3[0] = newgroup[group].pixM[0];
            pix3[1] = newgroup[group].pixM[1];
            pix3[2] = newgroup[group].pixM[2];

            mcount++;
         }
         
         if (mcount == 0) break;
      }
   }

   zfree(pixgroup);
   zfree(pixstack);
   zfree(groupcount);
   zfree(newgroup);

   return;
}


//  paint borders between the groups of contiguous pixels

void painting_paintborders()
{
   using namespace painting_names;

   int            ii, kk, px, py, cc;
   float          *pix3, *pixL, *pixA;
   
   if (! borders) return;
   
   cc = E3pxm->ww * E3pxm->hh;
   char * pixblack = zmalloc(cc,"painting");
   memset(pixblack,0,cc);

   if (sa_stat == 3)
   {
      for (ii = 0; ii < Fpxb->ww * Fpxb->hh; ii++)                         //  find pixels in select area
      {
         if (! sa_pixmap[ii]) continue;
         py = ii / Fpxb->ww;
         px = ii - py * Fpxb->ww;
         if (px < 1 || py < 1) continue;

         pix3 = PXMpix(E3pxm,px,py);
         pixL = PXMpix(E3pxm,px-1,py);
         pixA = PXMpix(E3pxm,px,py-1);
         
         if (pix3[0] != pixL[0] || pix3[1] != pixL[1] || pix3[2] != pixL[2])
         {
            kk = ii - 1;
            if (pixblack[kk]) continue;
            kk += 1;
            pixblack[kk] = 1;
            continue;
         }

         if (pix3[0] != pixA[0] || pix3[1] != pixA[1] || pix3[2] != pixA[2])
         {
            kk = ii - E3pxm->ww;
            if (pixblack[kk]) continue;
            kk += E3pxm->ww;
            pixblack[kk] = 1;
         }
      }

      for (ii = 0; ii < Fpxb->ww * Fpxb->hh; ii++)                         //  find pixels in select area
      { 
         if (! sa_pixmap[ii]) continue;
         py = ii / Fpxb->ww;
         px = ii - py * Fpxb->ww;
         if (px < 1 || py < 1) continue;

         if (! pixblack[ii]) continue;
         pix3 = PXMpix(E3pxm,px,py);
         pix3[0] = pix3[1] = pix3[2] = 0;
      }
   }
         
   else
   {
      for (py = 1; py < E3pxm->hh; py++)                                   //  loop all pixels
      for (px = 1; px < E3pxm->ww; px++)                                   //  omit top and left
      {
         pix3 = PXMpix(E3pxm,px,py);                                       //  output pixel
         pixL = PXMpix(E3pxm,px-1,py);                                     //  pixel to left
         pixA = PXMpix(E3pxm,px,py-1);                                     //  pixel above
         
         if (pix3[0] != pixL[0] || pix3[1] != pixL[1] || pix3[2] != pixL[2])
         {
            kk = E3pxm->ww * py + px-1;                                    //  have horiz. transition
            if (pixblack[kk]) continue;
            kk += 1;
            pixblack[kk] = 1;
            continue;
         }

         if (pix3[0] != pixA[0] || pix3[1] != pixA[1] || pix3[2] != pixA[2])
         {
            kk = E3pxm->ww * (py-1) + px;                                  //  have vertical transition
            if (pixblack[kk]) continue;
            kk += E3pxm->ww;
            pixblack[kk] = 1;
         }
      }

      for (py = 1; py < E3pxm->hh; py++)
      for (px = 1; px < E3pxm->ww; px++)
      {
         kk = E3pxm->ww * py + px;
         if (! pixblack[kk]) continue;
         pix3 = PXMpix(E3pxm,px,py);
         pix3[0] = pix3[1] = pix3[2] = 0;
      }
   }
   
   zfree(pixblack);
   return;
}


//  blend edges of selected area

void painting_blend()
{
   int         ii, px, py, dist;
   float       *pix1, *pix3;
   double      f1, f2;
   
   if (sa_stat == 3 && sa_blend > 0)
   {
      for (ii = 0; ii < Fpxb->ww * Fpxb->hh; ii++)                         //  find pixels in select area
      {
         dist = sa_pixmap[ii];
         if (! dist || dist >= sa_blend) continue;

         py = ii / Fpxb->ww;
         px = ii - py * Fpxb->ww;
         pix1 = PXMpix(E1pxm,px,py);                                       //  input pixel
         pix3 = PXMpix(E3pxm,px,py);                                       //  output pixel

         f2 = 1.0 * dist / sa_blend;                                       //  changes over distance sa_blend
         f1 = 1.0 - f2;

         pix3[0] = f1 * pix1[0] + f2 * pix3[0];
         pix3[1] = f1 * pix1[1] + f2 * pix3[1];
         pix3[2] = f1 * pix1[2] + f2 * pix3[2];
      }
   }

   return;
}


/**************************************************************************/

//  False Colors menu function
//  Gradually shift selected RGB colors into other colors.

editfunc    EFfalscolr;                                                    //  edit function data
float       falsred, falsgreen, falsblue;

void m_false_color(GtkWidget *, const char *)
{
   int    falscolr_dialog_event(zdialog* zd, const char *event);
   void * falscolr_thread(void *);

   F1_help_topic = "false_colors";
   
   EFfalscolr.funcname = "false_color";                                    //  function name
   EFfalscolr.Fprev = 1;                                                   //  use preview
   EFfalscolr.Farea = 2;                                                   //  select area 0/1/2 = delete/ignore/usable
   EFfalscolr.threadfunc = falscolr_thread;                                //  thread function

   if (! edit_setup(EFfalscolr)) return;                                   //  setup edit

/***
                  False Colors
        
      red:    green =======[]======= blue
      green:   blue ==========[]==== red
      blue:     red ======[]======== green

                          [done] [cancel]
***/

   zdialog *zd = zdialog_new(ZTX("False Colors"),mWin,Bdone,Bcancel,null);
   EFfalscolr.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"homog|space=3");
   zdialog_add_widget(zd,"vbox","vb3","hb1",0,"homog|space=3|expand");
   zdialog_add_widget(zd,"vbox","vb4","hb1",0,"homog|space=3");
   zdialog_add_widget(zd,"label","labg","vb2",ZTX("Red:     Green"));
   zdialog_add_widget(zd,"label","labb","vb2",ZTX("Green:    Blue"));
   zdialog_add_widget(zd,"label","labr","vb2",ZTX("Blue:      Red"));
   zdialog_add_widget(zd,"hscale","red","vb3","0|1|0.01|0.5");
   zdialog_add_widget(zd,"hscale","green","vb3","0|1|0.01|0.5");
   zdialog_add_widget(zd,"hscale","blue","vb3","0|1|0.01|0.5");
   zdialog_add_widget(zd,"label","labb","vb4",Bblue);
   zdialog_add_widget(zd,"label","labr","vb4",Bred);
   zdialog_add_widget(zd,"label","labg","vb4",Bgreen);
   
   falsred = falsgreen = falsblue = 0.5;

   zdialog_resize(zd,300,0);
   zdialog_run(zd,falscolr_dialog_event,"save");                           //  run dialog - parallel
   return;
}


//  false color dialog event and completion function

int falscolr_dialog_event(zdialog *zd, const char *event)
{
   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();                                     //  done
      else edit_cancel();                                                  //  cancel or destroy
      return 1;
   }

   if (strEqu(event,"apply")) {                                            //  from f_save()                   v.13.04
      edit_fullsize();
      return 1;
   }
   
   if (strEqu(event,"red")) 
      zdialog_fetch(zd,"red",falsred);

   if (strEqu(event,"green")) 
      zdialog_fetch(zd,"green",falsgreen);

   if (strEqu(event,"blue")) 
      zdialog_fetch(zd,"blue",falsblue);

   signal_thread();                                                        //  trigger update thread
   return 1;
}


//  thread function - launch multiple working threads to update image

void * falscolr_thread(void *)
{
   void  * falscolr_wthread(void *arg);                                    //  worker thread

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      if (sa_stat == 3) Fbusy_goal = sa_Npixel;                            //  set up progress monitor
      else  Fbusy_goal = E3pxm->ww * E3pxm->hh;
      Fbusy_done = 0;

      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(falscolr_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      Fbusy_goal = Fbusy_done = 0;
      CEF->Fmods++;                                                        //  image modified
      CEF->Fsaved = 0;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop warning
}


void * falscolr_wthread(void *arg)                                         //  worker thread function
{
   int         index = *((int *) (arg));
   int         px, py, ii, dist = 0;
   float       *pix1, *pix3;
   float       red1, green1, blue1, red3, green3, blue3;
   float       lossR, lossG, lossB, gainR, gainG, gainB;
   float       shift, move, maxrgb, f1, f2;

   for (py = index; py < E3pxm->hh; py += Nwt)                             //  loop all image pixels
   for (px = 0; px < E3pxm->ww; px++)
   {
      if (sa_stat == 3) {                                                  //  select area active
         ii = py * E3pxm->ww + px;
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  outside pixel
      }
      
      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel

      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];
      
      lossR = lossG = lossB = 0;
      gainR = gainG = gainB = 0;
      
      shift = falsred;                                                     //  red shift: green <-- red --> blue
      if (shift < 0.5) {                                                   //  0.0 ... 0.5
         move = red1 * 2.0 * (0.5 - shift);
         lossR += move;
         gainG += move;
      }
      if (shift > 0.5) {                                                   //  0.5 ... 1.0
         move = red1 * 2.0 * (shift - 0.5);
         lossR += move;
         gainB += move;
      }

      shift = falsgreen;                                                   //  green shift: blue <-- green --> red
      if (shift < 0.5) {                                                   //  0.0 ... 0.5
         move = green1 * 2.0 * (0.5 - shift);
         lossG += move;
         gainB += move;
      }
      if (shift > 0.5) {                                                   //  0.5 ... 1.0
         move = green1 * 2.0 * (shift - 0.5);
         lossG += move;
         gainR += move;
      }

      shift = falsblue;                                                    //  blue shift: red <-- blue --> green
      if (shift < 0.5) {                                                   //  0.0 ... 0.5
         move = blue1 * 2.0 * (0.5 - shift);
         lossB += move;
         gainR += move;
      }
      if (shift > 0.5) {                                                   //  0.5 ... 1.0
         move = blue1 * 2.0 * (shift - 0.5);
         lossB += move;
         gainG += move;
      }

      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel
      
      red3 = red1 - lossR + gainR;
      green3 = green1 - lossG + gainG;
      blue3 = blue1 - lossB + gainB;
      
      maxrgb = red3;                                                       //  find max. new RGB color
      if (green3 > maxrgb) maxrgb = green3;
      if (blue3 > maxrgb) maxrgb = blue3;
      
      if (maxrgb > 255.9) {                                                //  if too big, rescale all colors
         red3 = red3 * 255.9 / maxrgb;
         green3 = green3 * 255.9 / maxrgb;
         blue3 = blue3 * 255.9 / maxrgb;
      }
      
      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         f1 = 1.0 * dist / sa_blend;                                       //    blend changes over sa_blend
         f2 = 1.0 - f1;
         red3 = f1 * red3 + f2 * red1;
         green3 = f1 * green3 + f2 * green1;
         blue3 = f1 * blue3 + f2 * blue1;
      }

      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************

   Vignette function
   
   1. Change the brightness from center to edge using a curve.
   2. Change the color from center to edge using a color and a curve.
      (the pixel varies between original RGB and selected color)
   3. Mouse click or drag on image sets a new vignette center.

***************************************************************************/

void  vign_mousefunc();

editfunc    EFvignette;
uint8       vignette_RGB[3] = { 0, 0, 255 };
int         vignette_spc;
float       vign_cx, vign_cy;
float       vign_rad;


void m_vignette(GtkWidget *, cchar *)
{
   int      Vign_dialog_event(zdialog *zd, cchar *event);
   void     Vign_curvedit(int);
   void *   Vign_thread(void *);

   F1_help_topic = "vignette";

   cchar       *title = ZTX("Vignette");

   EFvignette.funcname = "vignette";
   EFvignette.Farea = 2;                                                   //  select area usable
   EFvignette.Fprev = 1;                                                   //  use preview image
   EFvignette.threadfunc = Vign_thread;                                    //  thread function
   EFvignette.mousefunc = vign_mousefunc;                                  //  mouse function

   if (! edit_setup(EFvignette)) return;                                   //  setup edit

/***
            _____________________________
           |                             |
           |                             |
           |    curve drawing area       |
           |                             |
           |_____________________________| 
            center                   edge

            (o) Brightness  (o) Color [___]
            Curve File: [ Open ] [ Save ]

***/

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);
   EFvignette.zd = zd;

   zdialog_add_widget(zd,"frame","frame","dialog",0,"expand");
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labcenter","hb1",ZTX("center"),"space=4");
   zdialog_add_widget(zd,"label","space","hb1",0,"expand");
   zdialog_add_widget(zd,"label","labedge","hb1",ZTX("edge"),"space=5");

   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=3");
   zdialog_add_widget(zd,"radio","RBbrite","hb2",Bbrightness,"space=5");
   zdialog_add_widget(zd,"radio","RBcolor","hb2",Bcolor,"space=5");
   zdialog_add_widget(zd,"colorbutt","color","hb2","0|0|255");

   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labcurve","hb3",Bcurvefile,"space=5");
   zdialog_add_widget(zd,"button","load","hb3",Bopen,"space=5");
   zdialog_add_widget(zd,"button","savecurve","hb3",Bsave,"space=5");
   
   vignette_RGB[0] = vignette_RGB[1] = 0;                                  //  initial color = blue
   vignette_RGB[2] = 255;
   
   vign_cx = E3pxm->ww / 2;                                                //  initial vignette center
   vign_cy = E3pxm->hh / 2;
   
   vign_rad = vign_cx * vign_cx + vign_cy * vign_cy;                       //  radius = distance to corners
   vign_rad = sqrtf(vign_rad);
   
   zdialog_stuff(zd,"RBbrite",1);                                          //  default curve = brightness

   GtkWidget *frame = zdialog_widget(zd,"frame");                          //  set up curve edit
   spldat *sd = splcurve_init(frame,Vign_curvedit);
   EFvignette.curves = sd;

   sd->vert[0] = 0;                                                        //  curve 0 = active curve = brightness
   sd->nap[0] = 2;
   sd->apx[0][0] = 0.01;
   sd->apy[0][0] = 0.5;
   sd->apx[0][1] = 0.99;
   sd->apy[0][1] = 0.5;
   splcurve_generate(sd,0);                                                //  generate curve data

   sd->vert[1] = 0;                                                        //  curve 1 = brightness curve
   sd->nap[1] = 2;
   sd->apx[1][0] = 0.01;
   sd->apy[1][0] = 0.5;
   sd->apx[1][1] = 0.99;
   sd->apy[1][1] = 0.5;
   splcurve_generate(sd,1);

   sd->vert[2] = 0;                                                        //  curve 2 = color curve
   sd->nap[2] = 2;
   sd->apx[2][0] = 0.01;
   sd->apy[2][0] = 0.01;
   sd->apx[2][1] = 0.99;
   sd->apy[2][1] = 0.01;
   splcurve_generate(sd,2);

   sd->Nspc = 1;                                                           //  only one active curve
   vignette_spc = 1;                                                       //  initial curve = brightness

   zdialog_resize(zd,300,300);
   zdialog_run(zd,Vign_dialog_event,"save");                               //  run dialog - parallel
   
   takeMouse(vign_mousefunc,dragcursor);                                   //  connect mouse function

   return;
}


//  dialog event and completion callback function

int Vign_dialog_event(zdialog *zd, cchar *event)
{
   void     Vign_curvedit(int);

   spldat   *sd = EFvignette.curves;
   int      ii, jj;
   char     color[20];
   cchar    *pp;

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) {                                                //  done
         float R = 1.0 * E0pxm->ww / E3pxm->ww;
         vign_cx = R * vign_cx;                                            //  scale geometries to full size
         vign_cy = R * vign_cy;
         vign_rad = R * vign_rad;
         edit_done();
      }
      else edit_cancel();                                                  //  cancel
      return 1;
   }

   if (strEqu(event,"save")) {                                             //  from f_save()                   v.13.04
      float R = 1.0 * E0pxm->ww / E3pxm->ww;
      vign_cx = R * vign_cx;                                               //  scale geometries to full size   v.13.05
      vign_cy = R * vign_cy;
      vign_rad = R * vign_rad;
      edit_fullsize();
      return 1;
   }
   
   if (strEqu(event,"focus"))                                              //  toggle mouse capture
      takeMouse(vign_mousefunc,dragcursor);                                //  connect mouse function

   if (strnEqu(event,"RB",2)) {                                            //  new choice of curve
      ii = strcmpv(event,"RBbrite","RBcolor",null);
      vignette_spc = ii;
      sd->nap[0] = sd->nap[ii];                                            //  copy new curve to curve 0
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }
      splcurve_generate(sd,0);                                             //  regenerate curve 0
      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve
      signal_thread();                                                     //  v.13.04
   }

   if (strEqu(event,"blendwidth")) signal_thread();

   if (strEqu(event,"color")) {                                            //  change color
      zdialog_fetch(zd,"color",color,19);                                  //  get color from color wheel
      pp = strField(color,"|",1);
      if (pp) vignette_RGB[0] = atoi(pp);
      pp = strField(color,"|",2);
      if (pp) vignette_RGB[1] = atoi(pp);
      pp = strField(color,"|",3);
      if (pp) vignette_RGB[2] = atoi(pp);
      signal_thread();                                                     //  trigger update thread
   }

   if (strEqu(event,"load")) {                                             //  load saved curve 
      splcurve_load(sd);
      Vign_curvedit(0);
      signal_thread();
      return 0;
   }

   if (strEqu(event,"savecurve")) {                                        //  save curve to file 
      splcurve_save(sd);
      return 0;
   }
   
   return 0;
}


//  get mouse position and set new center for vignette

void vign_mousefunc()                                                      //  mouse function
{
   if (! LMclick && ! Mdrag) return;
   LMclick = 0;
   
   vign_cx = Mxposn;                                                       //  new vignette center = mouse position
   vign_cy = Myposn;

   Mxdrag = Mydrag = 0;

   signal_thread();                                                        //  trigger image update
   return;
}


//  this function is called when the curve is edited

void Vign_curvedit(int)
{
   int      ii = vignette_spc, jj;
   spldat   *sd = EFvignette.curves;

   sd->nap[ii] = sd->nap[0];                                               //  copy curve 0 to current curve
   for (jj = 0; jj < sd->nap[0]; jj++) {
      sd->apx[ii][jj] = sd->apx[0][jj];
      sd->apy[ii][jj] = sd->apy[0][jj];
   }
   for (jj = 0; jj < 1000; jj++)
      sd->yval[ii][jj] = sd->yval[0][jj];

   signal_thread();                                                        //  update image
   return;
}


//  thread function

void * Vign_thread(void *)
{
   void * Vign_wthread(void *arg);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      for (int ii = 0; ii < Nwt; ii++)                                     //  start working threads
         start_wthread(Vign_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion
      
      CEF->Fmods++; 
      CEF->Fsaved = 0;
      mwpaint2(); 
   }

   return 0;                                                               //  not executed, stop g++ warning
}


//  working thread

void * Vign_wthread(void *arg)
{
   float       *pix1, *pix3;
   int         index, ii, kk, px, py, dist = 0;
   float       cx, cy, rad, radx, rady, f1, f2, xval, yval;
   float       red1, green1, blue1, red3, green3, blue3, cmax;
   spldat      *sd = EFvignette.curves;

   cx = vign_cx;                                                           //  vignette center (mouse)
   cy = vign_cy;

   index = *((int *) arg);

   for (py = index; py < E3pxm->hh; py += Nwt)                             //  loop all image pixels
   for (px = 0; px < E3pxm->ww; px++)
   {
      ii = py * E3pxm->ww + px;

      if (sa_stat == 3) {                                                  //  select area active 
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel is outside area
      }
      
      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel

      red1 = pix1[0];                                                      //  input RGB
      green1 = pix1[1];
      blue1 = pix1[2];

      radx = px - cx;                                                      //  distance from vignette center
      rady = py - cy;
      rad = sqrtf(radx*radx + rady*rady);                                  //  (px,py) distance from center
     
      xval = rad / vign_rad;                                               //  scale 0 to 1.0
      kk = 999.0 * xval;                                                   //  scale 0 to 999
      if (kk > 999) kk = 999;                                              //  beyond radius

      yval = sd->yval[1][kk];                                              //  brightness curve y-value 0 to 1.0
      if (yval > 1.0) yval = 1.0;
      yval = 2.0 * yval;                                                   //  0 to 2.0
      
      red3 = yval * red1;                                                  //  adjust brightness
      green3 = yval * green1;
      blue3 = yval * blue1;

      yval = sd->yval[2][kk];                                              //  color curve y-value 0 to 1.0
      if (yval > 1.0) yval = 1.0;
      f1 = yval;                                                           //  0 to 1.0   new color
      f2 = 1.0 - f1;                                                       //  1.0 to 0   old color

      red3 = f1 * vignette_RGB[0] + f2 * red3;                             //  mix input and vignette color
      green3 = f1 * vignette_RGB[1] + f2 * green3;
      blue3 = f1 * vignette_RGB[2] + f2 * blue3;

      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         f1 = 1.0 * dist / sa_blend;                                       //    blend changes over sa_blend
         f2 = 1.0 - f1;
         red3 = f1 * red3 + f2 * red1;
         green3 = f1 * green3 + f2 * green1;
         blue3 = f1 * blue3 + f2 * blue1;
      }
      
      cmax = red3;                                                         //  detect overflow
      if (green3 > cmax) cmax = green3;
      if (blue3 > cmax) cmax = blue3;
      
      if (cmax > 255.9) {                                                  //  stop overflow
         red3 = red3 * 255.9 / cmax;
         green3 = green3 * 255.9 / cmax;
         blue3 = blue3 * 255.9 / cmax;
      }

      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, stop g++ warning
}



