/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007 2008 2009 2010 2011 2012 2013 Michael Cornelison
   Source URL: http://kornelix.com/fotoxx
   Contact: kornelix@gmx.de
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************/

#define EX extern                                                          //  enable extern declarations
#include "fotoxx.h"

/**************************************************************************
   Fotoxx image edit - Image menu functions
***************************************************************************/


/**************************************************************************/

//  Rotate image right or left 90 degrees to upright the image.
//  This is not an edit transaction: file is replaced and re-opened.

zdialog  *zdturn;

void m_rotate90(GtkWidget *, cchar *menu)                                  //  v.13.08
{
   int  turnevent(zdialog *, cchar *);
   void turnfunc(GtkWidget *, cchar *menu);

   if (! curr_file) return;                                                //  v.13.09
   F1_help_topic = "rotate";

   if (mod_keep()) return;                                                 //  unsaved edits
   if (! menulock(1)) return;

   zdturn = zdialog_new(ZTX("Rotate Image 90º"),mWin,null);
   zdialog_add_widget(zdturn,"vbox","vb1","dialog",0,"space=5");
   GtkWidget *tbar = create_toolbar(zdialog_widget(zdturn,"vb1"),32);
   gtk_toolbar_set_style(GTK_TOOLBAR(tbar),GTK_TOOLBAR_BOTH);
   add_toolbar_button(tbar,"left",ZTX("rotate -90º"),"rotate-left.png",turnfunc);
   add_toolbar_button(tbar,"right",ZTX("rotate +90º"),"rotate-right.png",turnfunc);
   add_toolbar_button(tbar,"cancel",0,"gtk-cancel",turnfunc);
   zdialog_resize(zdturn,200,0);
   zdialog_run(zdturn,turnevent,"save");
   return;
}


int turnevent(zdialog *zd, cchar *event)
{
   if (zdturn->zstat) {                                                    //  [x] button
      zdialog_free(zdturn);
      menulock(0);
   }
   return 1;
}


void turnfunc(GtkWidget *, cchar *menu) 
{
   int      angle = 0;
   
   if (strEqu(menu,"cancel")) {
      zdialog_free(zdturn);
      menulock(0);
      return;
   }
   
   if (strEqu(menu,"right")) angle = +90;
   if (strEqu(menu,"left")) angle = -90;
   
   mutex_lock(&Fpixmap_lock);                                              //  block window updates
   E0pxm = PXM_load(curr_file,1);                                          //  load poss. 16-bit image         v.12.10
   if (E0pxm) {
      E3pxm = PXM_rotate(E0pxm,angle);                                     //  rotate
      PXM_free(E0pxm);
      E0pxm = E3pxm;
      E3pxm = 0;
   }
   mutex_unlock(&Fpixmap_lock);

   if (strEqu(curr_file_type,"other"))                                     //  if unsupported type, use jpg
      strcpy(curr_file_type,"jpg");

   f_save(curr_file,curr_file_type,curr_file_bpc);                         //  save file with curr. edits applied
   strcpy(curr_file_type,f_save_type);                                     //  update curr_file_xxx from f_save_xxx
   curr_file_size = f_save_size;

   f_open(curr_file,0);                                                    //  re-open
   return;
}


/**************************************************************************/

//  rotate image through any arbitrary angle

float       rotate_goal, rotate_angle;                                     //  target and actual rotation
int         rotate_resumetrim;
void        rotate_mousefunc();

editfunc    EFrotate;

void m_rotate(GtkWidget *, cchar *menu)                                    //  menu function    overhauled  v.13.08
{
   int    rotate_dialog_event(zdialog *zd, cchar *event);
   void * rotate_thread(void *);
   
   cchar  *rotmess = ZTX("Drag right edge with mouse");

   F1_help_topic = "rotate";

   EFrotate.funcname = "rotate";
   EFrotate.Fprev = 1;                                                     //  use preview
   EFrotate.Frestart = 1;                                                  //  allow restart                v.13.04
   EFrotate.threadfunc = rotate_thread;                                    //  thread function
   EFrotate.mousefunc = rotate_mousefunc;                                  //  mouse function
   if (! edit_setup(EFrotate)) return;                                     //  setup edit
   
   if (menu && strEqu(menu,"trim")) rotate_resumetrim = 1;                 //  resume trim after rotate
   else rotate_resumetrim = 0;

   rotate_goal = rotate_angle = 0;

/**
             ___________________________________
            |                                   |
            |    left [-90]  right [+90]        |
            |                                   |
            |    Rotate [0.0 |-+] degrees       |
            |    Drag right edge with mouse     |
            |    [Grid]                         |
            |                                   |
            |               [done] [cancel]     |
            |___________________________________|

**/

   zdialog *zd = zdialog_new(ZTX("Rotate Image"),mWin,Bdone,Bcancel,null);
   EFrotate.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog");                           //  v.13.08
   zdialog_add_widget(zd,"label","labl","hb1",ZTX("left"),"space=5");
   zdialog_add_widget(zd,"button","-90","hb1","-90");
   zdialog_add_widget(zd,"label","space","hb1",0,"space=10");
   zdialog_add_widget(zd,"label","labr","hb1",ZTX("right"),"space=5");
   zdialog_add_widget(zd,"button","+90","hb1","+90");

   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labrot","hb2",Brotate,"space=5");
   zdialog_add_widget(zd,"spin","degrees","hb2","-360|360|0.1|0");
   zdialog_add_widget(zd,"label","labdeg","hb2",ZTX("degrees"),"space=5");

   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labrot","hb3",ZTX(rotmess),"space=5");

   zdialog_add_widget(zd,"hbox","hb4","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","grid","hb4",ZTX("Grid"),"space=5");

   zdialog_run(zd,rotate_dialog_event,"save");                             //  run dialog, parallel

   takeMouse(rotate_mousefunc,dragcursor);                                 //  connect mouse function
   currgrid = 1;                                                           //  use rotate grid

   return;
}


//  dialog event and completion callback function

int rotate_dialog_event(zdialog *zd, cchar *event)
{
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) {                                                //  done
         rotate_angle = 0;                                                 //  rotate full size image
         edit_done();
      }
      else edit_cancel();                                                  //  canceled

      currgrid = 0;                                                        //  restore normal grid settings
      mwpaint2();
      
      if (rotate_resumetrim) m_trim(0,0);                                  //  resume edit trim       v.12.06
      return 1;
   }

   if (strEqu(event,"save")) {                                             //  from f_save()          v.13.04
      rotate_angle = 0;
      edit_fullsize();
      CEF->Fmods = 0;                                                      //  v.13.08
      return 1;
   }
   
   if (strEqu(event,"restart")) {                                          //  new image              v.13.04
      rotate_goal = rotate_angle = 0;
      rotate_resumetrim = 0;
      zdialog_stuff(zd,"degrees",0);
   }

   if (strstr("+90 -90 degrees",event))                                    //  rotate action
   {
      if (strEqu(event,"+90"))                                             //  90 deg. CW
         rotate_goal += 90;

      if (strEqu(event,"-90"))                                             //  90 deg. CCW
         rotate_goal -= 90;

      if (strEqu(event,"degrees"))                                         //  degrees adjustment
         zdialog_fetch(zd,"degrees",rotate_goal);

      if (rotate_goal > 180) rotate_goal -= 360;                           //  keep within -180 to +180 deg.
      if (rotate_goal < -180) rotate_goal += 360;
      if (fabsf(rotate_goal) < 0.01) rotate_goal = 0;                      //  = 0 within my precision
      zdialog_stuff(zd,"degrees",rotate_goal);

      signal_thread();                                                     //  do rotate in thread
   }

   if (strEqu(event,"grid")) m_gridlines(0,"grid 1");                      //  grid settings dialog      v.13.03

   return 1;
}


//  rotate mouse function - drag right edge of image up/down for rotation

void rotate_mousefunc()
{
   static int     mpx0 = 0, mpy0 = 0;
   static int     mpy1, mpy2, dist;
   zdialog        *zd = EFrotate.zd;

   if (! Mxdrag && ! Mydrag) return;                                       //  no drag underway
   
   if (! E3pxm) return;                                                    //  v.12.10
   if (Mxdrag < 0.8 * E3pxm->ww) return;                                   //  not right edge of image

   if (Mxdown != mpx0 || Mydown != mpy0) {
      mpx0 = Mxdown;                                                       //  new drag started
      mpy0 = mpy1 = Mydown;
   }
   
   mpy2 = Mydrag;                                                          //  use drag as rotation 
   Mxdrag = Mydrag = 0;                                                    //  v.12.11

   dist = mpy2 - mpy1;                                                     //  drag distance
   mpy1 = mpy2;                                                            //  reset origin for next time
   if (! dist) return;

   rotate_goal += 30.0 * dist / E3pxm->ww;                                 //  convert radians to degrees, reduced

   if (rotate_goal > 180) rotate_goal -= 360;                              //  keep within -180 to +180 deg.
   if (rotate_goal < -180) rotate_goal += 360;
   if (fabsf(rotate_goal) < 0.01) rotate_goal = 0;                         //  = 0 within my precision
   zdialog_stuff(zd,"degrees",rotate_goal);

   signal_thread();                                                        //  do rotate in thread
   return;
}


//  rotate thread function

void * rotate_thread(void *)
{
   int         px3, py3, px9, py9;
   int         wwcut, hhcut, ww, hh;
   float       trim_angle, radians;
   float       *pix3, *pix9;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      while (rotate_goal != rotate_angle)
      {
         mutex_lock(&Fpixmap_lock);
         Ffuncbusy++;                                                      //  v.12.02
         
         if (rotate_goal == 0) {                                           //  no rotation
            PXM_free(E3pxm);                                               //  E1 >> E3
            E3pxm = PXM_copy(E1pxm);
            CEF->Fmods = 0;                                                //  v.13.04
            CEF->Fsaved = 0;
            rotate_angle = 0;
         }
         
         if (rotate_goal != rotate_angle)                                  //  rotate to goal angle
         {
            E9pxm = PXM_rotate(E1pxm,rotate_goal);                         //  E9 is rotated E1
            PXM_free(E3pxm);
            E3pxm = E9pxm;
            CEF->Fmods++; 
            CEF->Fsaved = 0;
            E9pxm = 0;
            rotate_angle = rotate_goal;

            trim_angle = fabsf(rotate_goal);
            while (trim_angle > 45) trim_angle -= 90;
            radians = fabsf(trim_angle / 57.296);
            wwcut = int(E3pxm->hh * sinf(radians) + 1);                    //  amount to trim
            hhcut = int(E3pxm->ww * sinf(radians) + 1);
            ww = E3pxm->ww - 2 * wwcut;
            hh = E3pxm->hh - 2 * hhcut;
            if (ww > 10 && hh > 10) {
               E9pxm = PXM_make(ww,hh);
               
               for (py3 = hhcut; py3 < E3pxm->hh-hhcut; py3++)             //  E9 = trimmed E3
               for (px3 = wwcut; px3 < E3pxm->ww-wwcut; px3++)
               {
                  px9 = px3 - wwcut;
                  py9 = py3 - hhcut;
                  pix3 = PXMpix(E3pxm,px3,py3);
                  pix9 = PXMpix(E9pxm,px9,py9);
                  pix9[0] = pix3[0];
                  pix9[1] = pix3[1];
                  pix9[2] = pix3[2];
               }

               PXM_free(E3pxm);                                            //  E3 = E9
               E3pxm = E9pxm;
               E9pxm = 0;
            }

            CEF->Fmods++; 
            CEF->Fsaved = 0;
         }
         
         mutex_unlock(&Fpixmap_lock);                                      //  v.12.06
         Ffuncbusy--;
         mwpaint2();
         zsleep(0.1);
      }
   }

   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************/

//  trim (crop) image - use mouse to select image region to retain

int      trimx1, trimy1, trimx2, trimy2;                                   //  current trim rectangle
int      trimpx1, trimpy1, trimpx2, trimpy2;                               //  prior trim rectangle
float    trimR;                                                            //  trim ratio, width/height
int      trimportrait = 0;                                                 //  0 = landscape, 1 = portrait

void  trim_initz(int autotrim);
void  trim_mousefunc();                                                    //  edit trim margins with mouse
void  trim_dialog();
void  trim_customize();
void  trim_trim(int mode);                                                 //  show trim area, trim image

editfunc    EFtrim;


void m_trim(GtkWidget *, cchar *menu)
{
   int      ii, init;
   char     *pp;
   cchar    *defbutts[6] = { "1:1", "2:1", "3:2", "4:3", "16:9", "gold" };

   F1_help_topic = "trim_image";

   EFtrim.funcname = "trim";
   EFtrim.Frestart = 1;                                                    //  allow restart                v.13.04
   EFtrim.mousefunc = trim_mousefunc;
   if (! edit_setup(EFtrim)) return;                                       //  setup edit
   
   for (ii = 0; ii < 6; ii++)
   {                                                                       //  initialize buttons           v.13.04
      init = 0;
      pp = trimbuttons[ii];
      if (! pp || ! *pp || strEqu(pp,"undefined")) init = 1;
      pp = trimratios[ii];
      if (! pp || ! *pp || ! strchr(pp,':')) init = 1;
      if (init) {
         trimbuttons[ii] = strdupz(defbutts[ii],20,"trim");
         trimratios[ii] =  strdupz(defbutts[ii],20,"trim");
         if (ii == 5) strcpy(trimratios[ii],"1.62:1");
      }
   }

   if (menu && strEqu(menu,"autotrim"))                                    //  autotrim: keep preset trim margins
      trim_initz(1);
   else trim_initz(0);

   return;
}


//  initialize trim for new image

void trim_initz(int autotrim)                                              //  v.13.04
{
   int      xmargin, ymargin;
   
   trimpx1 = 0;                                                            //  set prior trim rectangle
   trimpx2 = Fpxb->ww;                                                     //    = 100% of image
   trimpy1 = 0;
   trimpy2 = Fpxb->hh;
   
   if (autotrim) {                                                         //  autotrim: keep preset trim rectangle
      trimsize[0] = trimx2 - trimx1;                                       //  else initz. using prior trim size
      trimsize[1] = trimy2 - trimy1;
   }

   if (trimsize[0] > Fpxb->ww) trimsize[0] = Fpxb->ww;                     //  if > image size, shrink
   if (trimsize[1] > Fpxb->hh) trimsize[1] = Fpxb->hh;

   if (trimsize[0] < 0.2 * Fpxb->ww) trimsize[0] = 0.2 * Fpxb->ww;         //  if < 20% image size, expand
   if (trimsize[1] < 0.2 * Fpxb->hh) trimsize[1] = 0.2 * Fpxb->hh;

   xmargin = 0.5 * (Fpxb->ww - trimsize[0]);                               
   ymargin = 0.5 * (Fpxb->hh - trimsize[1]);
   trimx1 = xmargin;
   trimx2 = Fpxb->ww - xmargin;
   trimy1 = ymargin;
   trimy2 = Fpxb->hh - ymargin;
   
   trimR = 1.0 * trimsize[0] / trimsize[1];                                //  trim ratio = width/height

   trim_trim(0);                                                           //  show trim area in image
   trim_dialog();                                                          //  start dialog
   return;
}


//  dialog function is called from two places

void trim_dialog()
{
   int    trim_dialog_event(zdialog *zd, cchar *event);

   cchar       *trim_message = ZTX("Drag middle to move, drag corners to resize.");
   char        text[20];
   int         ii;
   zdialog     *zd;

/**
       ________________________________________________________
      |                                                        |
      |    Drag middle to move, drag corners to resize         |
      |                                                        |
      |   width [____|-|+]  height [____|-|+]  ratio [____]    |           //  width/height inputs
      |   [1:1] [2:1] [3:2] [4:3] [16:9] [gold] [customize]    |
      |   (o) portrait  (o) landscape    [x] Lock Ratio        |
      |                                                        |
      |                    [Rotate] [Apply] [Done] [Cancel]    |
      |________________________________________________________|

**/
   
   F1_help_topic = "trim_image";
   
   if (! EFtrim.zd)                                                        //  v.13.04
   {
      zd = zdialog_new(ZTX("Trim/Crop"),mWin,Brotate,Bapply,Bdone,Bcancel,null);
      EFtrim.zd = zd;

      zdialog_add_widget(zd,"label","lab1","dialog",trim_message,"space=8");
      zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=4");
      zdialog_add_widget(zd,"label","labW","hb1",ZTX("width"),"space=3");
      zdialog_add_widget(zd,"spin","width","hb1","20|9999|1|1000");
      zdialog_add_widget(zd,"label","space","hb1",0,"space=5");
      zdialog_add_widget(zd,"label","labH","hb1",ZTX("height"),"space=3");
      zdialog_add_widget(zd,"spin","height","hb1","20|9999|1|600");
      zdialog_add_widget(zd,"label","space","hb1",0,"space=5");
      zdialog_add_widget(zd,"label","labR","hb1",ZTX("ratio"),"space=3");
      zdialog_add_widget(zd,"label","ratio","hb1","1.67   ");
      zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=4");                  //  ratio buttons inserted here
      zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=4");
      zdialog_add_widget(zd,"radio","portrait","hb3",ZTX("portrait"),"space=3");
      zdialog_add_widget(zd,"radio","landscape","hb3",ZTX("landscape"),"space=3");
      zdialog_add_widget(zd,"check","lock","hb3",ZTX("Lock Ratio"),"space=15");

      for (ii = 0; ii < 6; ii++)                                                 //  insert custom ratio buttons
         zdialog_add_widget(zd,"button",trimbuttons[ii],"hb2",trimbuttons[ii]);
      zdialog_add_widget(zd,"button","customize","hb2",ZTX("customize"));        //  and [customize] button
   }
   
   zd = EFtrim.zd;

   zdialog_stuff(zd,"width",trimsize[0]);                                  //  stuff width, height, ratio
   zdialog_stuff(zd,"height",trimsize[1]);
   sprintf(text,"%.2f  ",trimR);
   zdialog_stuff(zd,"ratio",text);

   zdialog_stuff(zd,"portrait",trimportrait);                              //  preserve portrait/landscape option
   zdialog_stuff(zd,"landscape",1-trimportrait);                           //  v.13.08
   
   zdialog_run(zd,trim_dialog_event,"save");                               //  run dialog - parallel
   takeMouse(trim_mousefunc,dragcursor);                                   //  connect mouse function 
   
   return;
}


//  dialog event and completion callback function

int trim_dialog_event(zdialog *zd, cchar *event)                           //  overhauled
{
   static int  flip = 0;
   int         width, height, delta;
   int         ii, rlock;
   float       r1, r2, ratio = 0;
   char        text[20];
   cchar       *pp;
   
   if (zd->zstat)                                                          //  dialog complete
   {
      draw_toplines(2);                                                    //  erase rectangle outline
      erase_topcircles();                                                  //  and target circle               v.12.09
      
      if (zd->zstat == 1) {                                                //  rotate                          v.12.06
         edit_cancel();
         m_rotate(0,"trim");                                               //  do rotate and come back
         return 1;
      }
      
      if (zd->zstat == 2) {                                                //  apply                           v.13.04
         zd->zstat = 0;                                                    //  keep dialog active
         trim_trim(1);
         trim_initz(0);
         Fzoom = 0;
         return 1;
      }
      
      if (zd->zstat == 3) {                                                //  done
         trim_trim(1);                                                     //  do trim on image
         Fzoom = 0;
         edit_done();
         return 1;
      }

      edit_cancel();                                                       //  cancel or kill
      return 1;
   }

   if (strEqu(event,"focus")) {
      takeMouse(trim_mousefunc,dragcursor);                                //  connect mouse function          v.12.01
      return 1;
   }
   
   if (strEqu(event,"save")) {                                             //  from f_save()                   v.13.04
      trim_trim(1);                                                        //  do trim on image
      trim_initz(0);
      Fzoom = 0;
      return 1;
   }
   
   if (strEqu(event,"restart")) {                                          //  new image                       v.13.04
      trim_initz(0);
      return 1;
   }
   
   if (strEqu(event,"customize")) {
      freeMouse();
      trim_customize();                                                    //  do customize dialog
      takeMouse(trim_mousefunc,dragcursor); 
      if (! EFtrim.zd) return 1;
      zdialog_free(EFtrim.zd);
      EFtrim.zd = 0;
      trim_initz(0);                                                       //  v.13.04
      return 1;
   }

   if (strstr("width height",event))                                       //  get direct width/height inputs
   {
      zdialog_fetch(zd,"width",width);
      zdialog_fetch(zd,"height",height);
      zdialog_fetch(zd,"lock",rlock);                                      //  lock ratio on/off
      
      if (strEqu(event,"width")) {
         if (width > Fpxb->ww) width = Fpxb->ww;
         if (rlock) {                                                      //  ratio locked
            height = width / trimR + 0.5;                                  //  try to keep ratio
            if (height > Fpxb->hh) height = Fpxb->hh;
            zdialog_stuff(zd,"height",height);
         }
      }

      if (strEqu(event,"height")) {
         if (height > Fpxb->hh) height = Fpxb->hh;
         if (rlock) {
            width = height * trimR + 0.5;
            if (width > Fpxb->ww) width = Fpxb->ww;
            zdialog_stuff(zd,"width",width);
         }
      }

      flip = 1 - flip;                                                     //  alternates 0, 1, 0, 1 ...

      delta = width - trimsize[0];

      if (delta > 0) {                                                     //  increase width
         trimx1 = trimx1 - delta / 2;                                      //  left and right sides equally
         trimx2 = trimx2 + delta / 2;
         if (delta % 2) {                                                  //  if increase is odd
            trimx1 = trimx1 - flip;                                        //    add 1 alternatively to each side
            trimx2 = trimx2 + 1 - flip;
         }
         if (trimx1 < 0) {                                                 //  add balance to upper limit   v.13.06
            trimx2 = trimx2 - trimx1;
            trimx1 = 0;
         }
         if (trimx2 > Fpxb->ww) {                                          //  add balance to lower limit   v.13.06
            trimx1 = trimx1 - trimx2 + Fpxb->ww;
            trimx2 = Fpxb->ww;
         }
      }
      
      if (delta < 0) {                                                     //  decrease width
         trimx1 = trimx1 - delta / 2;
         trimx2 = trimx2 + delta / 2;
         if (delta % 2) {
            trimx1 = trimx1 + flip;
            trimx2 = trimx2 - 1 + flip;
         }
      }

      delta = height - trimsize[1];

      if (delta > 0) {                                                     //  increase height
         trimy1 = trimy1 - delta / 2;                                      //  top and bottom sides equally
         trimy2 = trimy2 + delta / 2;
         if (delta % 2) {                                                  //  if increase is odd
            trimy1 = trimy1 - flip;                                        //    add 1 alternatively to each side
            trimy2 = trimy2 + 1 - flip;
         }
         if (trimy1 < 0) {
            trimy2 = trimy2 - trimy1;
            trimy1 = 0;
         }
         if (trimy2 > Fpxb->hh) {
            trimy1 = trimy1 - trimy2 + Fpxb->hh;
            trimy2 = Fpxb->hh;
         }
      }
      
      if (delta < 0) {                                                     //  decrease height
         trimy1 = trimy1 - delta / 2;
         trimy2 = trimy2 + delta / 2;
         if (delta % 2) {
            trimy1 = trimy1 + flip;
            trimy2 = trimy2 - 1 + flip;
         }
      }

      if (trimx1 < 0) trimx1 = 0;                                          //  keep within limits
      if (trimx2 > Fpxb->ww) trimx2 = Fpxb->ww;                            //  use ww not ww-1              v.13.06
      if (trimy1 < 0) trimy1 = 0;
      if (trimy2 > Fpxb->hh) trimy2 = Fpxb->hh;
      
      width = trimx2 - trimx1;                                             //  new width and height
      height = trimy2 - trimy1;
      
      zdialog_stuff(zd,"width",width);                                     //  update dialog values
      zdialog_stuff(zd,"height",height);

      trimsize[0] = width;                                                 //  new trim rectangle
      trimsize[1] = height;

      if (! rlock)                                                         //  set new ratio if not locked
         trimR = 1.0 * trimsize[0] / trimsize[1];

      sprintf(text,"%.2f  ",trimR);                                        //  stuff new ratio
      zdialog_stuff(zd,"ratio",text);

      trim_trim(0);                                                        //  show trim area in image
      return 1;
   }

   for (ii = 0; ii < 6; ii++)                                              //  trim ratio buttons
      if (strEqu(event,trimbuttons[ii])) break;
   if (ii < 6) {
      r1 = r2 = ratio = 0;
      pp = strField(trimratios[ii],':',1);
      if (pp) r1 = atof(pp);
      pp = strField(trimratios[ii],':',2);
      if (pp) r2 = atof(pp);
      if (r1 > 0 && r2 > 0) ratio = r1/r2;
      if (ratio < 0.1 || ratio > 10) ratio = 1.0;
      if (! ratio) return 1;
      zdialog_stuff(zd,"lock",1);                                          //  assume lock is wanted
      trimR = ratio;
   }
   
   if (strstr("portrait landscape",event)) {                               //  v.13.08
      zdialog_fetch(zd,"portrait",trimportrait);
      ratio = trimR;
   }

   if (ratio)                                                              //  ratio was changed
   {
      if (trimportrait && ratio > 1) ratio = 1.0 / ratio;                  //  v.13.08
      if (! trimportrait && ratio < 1) ratio = 1.0 / ratio;

      trimR = ratio;

      if (trimx2 - trimx1 > trimy2 - trimy1)
         trimy2 = trimy1 + (trimx2 - trimx1) / trimR;                      //  adjust smaller dimension
      else 
         trimx2 = trimx1 + (trimy2 - trimy1) * trimR;

      if (trimx2 > Fpxb->ww) {                                             //  if off the right edge,
         trimx2 = Fpxb->ww;                                                //  adjust height
         trimy2 = trimy1 + (trimx2 - trimx1) / trimR;
      }

      if (trimy2 > Fpxb->hh) {                                             //  if off the bottom edge,
         trimy2 = Fpxb->hh;                                                //  adjust width
         trimx2 = trimx1 + (trimy2 - trimy1) * trimR;
      }

      trimsize[0] = trimx2 - trimx1;                                       //  new rectangle dimensions
      trimsize[1] = trimy2 - trimy1;

      zdialog_stuff(zd,"width",trimsize[0]);                               //  stuff width, height, ratio
      zdialog_stuff(zd,"height",trimsize[1]);
      sprintf(text,"%.2f  ",trimR);
      zdialog_stuff(zd,"ratio",text);

      trim_trim(0);                                                        //  show trim area in image
      return 1;
   }
   
   return 1;
}


//  dialog to get custom trim button names and corresponding ratios

void trim_customize()
{
   char        text[20], blab[8], rlab[8];
   float       r1, r2, ratio;
   cchar       *pp;
   int         ii, zstat;
   
   zdialog *zd = zdialog_new(ZTX("Trim Buttons"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hbb","dialog",0,"homog|space=5");
   zdialog_add_widget(zd,"hbox","hbr","dialog",0,"homog|space=5");
   
   strcpy(blab,"butt-0");
   strcpy(rlab,"ratio-0");
   
   zdialog_add_widget(zd,"label","label","hbb",ZTX("label"),"space=5");
   zdialog_add_widget(zd,"label","ratio","hbr",ZTX("ratio"),"space=5");
   
   for (ii = 0; ii < 6; ii++) 
   {
      blab[5] = '0' + ii;
      rlab[6] = '0' + ii;
      zdialog_add_widget(zd,"entry",blab,"hbb",trimbuttons[ii],"scc=6");
      zdialog_add_widget(zd,"entry",rlab,"hbr",trimratios[ii],"scc=6");
   }
   
   zdialog_run(zd,0,"mouse");                                              //  run dialog
   zstat = zdialog_wait(zd);

   if (zstat == 1)                                                         //  done
   {
      for (ii = 0; ii < 6; ii++)                                           //  get custom button names
      {
         blab[5] = '0' + ii;
         zdialog_fetch(zd,blab,text,12);
         strTrim2(text);
         if (! *text) continue;
         zfree(trimbuttons[ii]);
         trimbuttons[ii] = strdupz(text,0,"trim");
      }

      for (ii = 0; ii < 6; ii++)                                           //  get custom ratios
      {
         rlab[6] = '0' + ii;
         zdialog_fetch(zd,rlab,text,12);
         strTrim2(text);
         r1 = r2 = ratio = 0;
         pp = strField(text,':',1);
         if (! pp) continue;
         r1 = atof(pp);
         pp = strField(text,':',2);
         if (! pp) continue;
         r2 = atof(pp);
         if (r1 > 0 && r2 > 0) ratio = r1/r2;
         if (ratio < 0.1 || ratio > 10) continue;
         zfree(trimratios[ii]);
         trimratios[ii] = strdupz(text,0,"trim");
      }
   }

   zdialog_free(zd);                                                       //  kill dialog
   return;
}


//  trim mouse function

void trim_mousefunc()
{
   int         mpx, mpy, xdrag, ydrag, rlock;
   int         corner, chop, moveall = 0;
   int         dx, dy, dd, d1, d2, d3, d4;
   char        text[20];
   float       drr;
   zdialog     *zd = EFtrim.zd;
   
   if (LMclick || Mxdrag || Mydrag)                                        //  mouse click or drag
   {
      if (LMclick) {
         mpx = Mxclick;                                                    //  click
         mpy = Myclick;
         xdrag = ydrag = 0;
      }
      else {
         mpx = Mxdrag;                                                     //  drag
         mpy = Mydrag;
         xdrag = Mxdrag - Mxdown;
         ydrag = Mydrag - Mydown;
         Mxdown = Mxdrag;                                                  //  reset drag origin
         Mydown = Mydrag;
      }
      
      if (Mxdrag || Mydrag) {
         moveall = 1;
         dd = 0.1 * (trimx2 - trimx1);                                     //  test if mouse is in the broad
         if (mpx < trimx1 + dd) moveall = 0;                               //    middle of the rectangle
         if (mpx > trimx2 - dd) moveall = 0;
         dd = 0.1 * (trimy2 - trimy1);
         if (mpy < trimy1 + dd) moveall = 0;
         if (mpy > trimy2 - dd) moveall = 0;
      }

      if (moveall) {                                                       //  yes, move the whole rectangle
         trimx1 += xdrag;
         trimx2 += xdrag;
         trimy1 += ydrag;
         trimy2 += ydrag;
         corner = 0;
      }

      else {                                                               //  no, find closest corner
         dx = mpx - trimx1;
         dy = mpy - trimy1;
         d1 = sqrt(dx*dx + dy*dy);                                         //  distance from NW corner
         
         dx = mpx - trimx2;
         dy = mpy - trimy1;
         d2 = sqrt(dx*dx + dy*dy);
         
         dx = mpx - trimx2;
         dy = mpy - trimy2;
         d3 = sqrt(dx*dx + dy*dy);
         
         dx = mpx - trimx1;
         dy = mpy - trimy2;
         d4 = sqrt(dx*dx + dy*dy);
         
         corner = 1;                                                       //  NW
         dd = d1;
         if (d2 < dd) { corner = 2; dd = d2; }                             //  NE
         if (d3 < dd) { corner = 3; dd = d3; }                             //  SE
         if (d4 < dd) { corner = 4; dd = d4; }                             //  SW
         
         if (corner == 1) { trimx1 = mpx; trimy1 = mpy; }                  //  move this corner to mouse
         if (corner == 2) { trimx2 = mpx; trimy1 = mpy; }
         if (corner == 3) { trimx2 = mpx; trimy2 = mpy; }
         if (corner == 4) { trimx1 = mpx; trimy2 = mpy; }
      }
      
      if (trimx1 > trimx2-10) trimx1 = trimx2-10;                          //  sanity limits
      if (trimy1 > trimy2-10) trimy1 = trimy2-10;

      zdialog_fetch(zd,"lock",rlock);                                      //  w/h ratio locked
      if (rlock && corner) {
         if (corner < 3)
            trimy2 = trimy1 + 1.0 * (trimx2 - trimx1) / trimR;
         else
            trimy1 = trimy2 - 1.0 * (trimx2 - trimx1) / trimR;
      }

      chop = 0;
      if (trimx1 < 0) {                                                    //  look for off the edge
         trimx1 = 0; 
         chop = 1; 
      }

      if (trimx2 > Fpxb->ww) {                                             //  after corner move
         trimx2 = Fpxb->ww;
         chop = 2; 
      }

      if (trimy1 < 0) { 
         trimy1 = 0; 
         chop = 3; 
      }

      if (trimy2 > Fpxb->hh) { 
         trimy2 = Fpxb->hh;
         chop = 4; 
      }

      if (rlock && chop) {                                                 //  keep ratio if off edge
         if (chop < 3)
            trimy2 = trimy1 + 1.0 * (trimx2 - trimx1) / trimR;
         else
            trimx2 = trimx1 + 1.0 * (trimy2 - trimy1) * trimR;
      }

      if (trimx1 > trimx2-10) trimx1 = trimx2-10;                          //  sanity limits
      if (trimy1 > trimy2-10) trimy1 = trimy2-10;

      if (trimx1 < 0) trimx1 = 0;                                          //  keep within visible area 
      if (trimx2 > Fpxb->ww) trimx2 = Fpxb->ww;
      if (trimy1 < 0) trimy1 = 0;
      if (trimy2 > Fpxb->hh) trimy2 = Fpxb->hh;

      trimsize[0] = trimx2 - trimx1;                                       //  new rectangle dimensions
      trimsize[1] = trimy2 - trimy1;

      drr = 1.0 * trimsize[0] / trimsize[1];                               //  new w/h ratio
      if (! rlock) trimR = drr;

      zdialog_stuff(zd,"width",trimsize[0]);                               //  stuff width, height, ratio
      zdialog_stuff(zd,"height",trimsize[1]);
      sprintf(text,"%.2f  ",trimR);
      zdialog_stuff(zd,"ratio",text);
      
      trim_trim(0);                                                        //  show trim area in image
   }

   LMclick = Mxdrag = Mydrag = 0;                                          //  v.12.11
   return;
}


//  darken image pixels outside of current trim margins
//  messy logic: update pixmaps only for changed pixels to increase speed

void trim_trim(int mode)
{
   int      ox1, oy1, ox2, oy2;                                            //  outer trim rectangle
   int      nx1, ny1, nx2, ny2;                                            //  inner trim rectangle
   int      px, py, px1, py1, px2, py2;
   int      radius;
   float    *pix1, *pix3;
 
   if (mode == 1)                                                          //  do the final trim
   {
      mutex_lock(&Fpixmap_lock);

      trimsize[0] = trimx2 - trimx1;                                       //  final trim rectangle
      trimsize[1] = trimy2 - trimy1;

      PXM_free(E3pxm);
      E3pxm = PXM_make(trimsize[0],trimsize[1]);                           //  new pixmap with requested size
      
      for (py1 = trimy1; py1 < trimy2; py1++)                              //  copy pixels
      for (px1 = trimx1; px1 < trimx2; px1++)
      {
         px2 = px1 - trimx1;
         py2 = py1 - trimy1;
         pix1 = PXMpix(E1pxm,px1,py1);
         pix3 = PXMpix(E3pxm,px2,py2);
         pix3[0] = pix1[0];
         pix3[1] = pix1[1];
         pix3[2] = pix1[2];
      }

      PXM_free(E1pxm);                                                     //  sync E1 to trimmed E3        v.13.04
      E1pxm = PXM_copy(E3pxm);

      CEF->Fmods++; 
      CEF->Fsaved = 0;
      mutex_unlock(&Fpixmap_lock);
      mwpaintnow();                                                        //  update window                v.13.04
      return;
   }

   erase_topcircles();
   radius = 0.02 * (Fpxb->ww + Fpxb->hh);                                  //  erase target circle          v.12.09
   px = 0.5 * (trimpx1 + trimpx2);
   py = 0.5 * (trimpy1 + trimpy2);
   mwpaint3(px-radius,py-radius,px+radius,py+radius);

   if (trimx1 < 0) trimx1 = 0;                                             //  keep within visible area
   if (trimx2 > Fpxb->ww) trimx2 = Fpxb->ww;
   if (trimy1 < 0) trimy1 = 0;
   if (trimy2 > Fpxb->hh) trimy2 = Fpxb->hh;

   if (trimpx1 < trimx1) ox1 = trimpx1;                                    //  outer rectangle
   else ox1 = trimx1;
   if (trimpx2 > trimx2) ox2 = trimpx2;
   else ox2 = trimx2;
   if (trimpy1 < trimy1) oy1 = trimpy1;
   else oy1 = trimy1;
   if (trimpy2 > trimy2) oy2 = trimpy2;
   else oy2 = trimy2;

   if (trimpx1 > trimx1) nx1 = trimpx1;                                    //  inner rectangle
   else nx1 = trimx1;
   if (trimpx2 < trimx2) nx2 = trimpx2;
   else nx2 = trimx2;
   if (trimpy1 > trimy1) ny1 = trimpy1;
   else ny1 = trimy1;
   if (trimpy2 < trimy2) ny2 = trimpy2;
   else ny2 = trimy2;
   
   trimpx1 = trimx1;                                                       //  set prior trim rectangle 
   trimpx2 = trimx2;                                                       //    from current trim rectangle
   trimpy1 = trimy1;
   trimpy2 = trimy2;

   ox1 -= 2;                                                               //  expand outer rectangle    v.12.03
   oy1 -= 2;
   ox2 += 2;
   oy2 += 2; 
   nx1 += 2;                                                               //  reduce inner rectangle
   ny1 += 2;
   nx2 -= 2;
   ny2 -= 2;
   
   if (ox1 < 0) ox1 = 0;
   if (oy1 < 0) oy1 = 0;
   if (ox2 > Fpxb->ww) ox2 = Fpxb->ww;
   if (oy2 > Fpxb->hh) oy2 = Fpxb->hh;
   
   if (nx1 < ox1) nx1 = ox1;                                               //  v.13.04
   if (ny1 < oy1) ny1 = oy1;
   if (nx2 > ox2) nx2 = ox2;
   if (ny2 > oy2) ny2 = oy2;
   
   if (nx2 < nx1) nx2 = nx1;
   if (ny2 < ny1) ny2 = ny1;
   
   for (py = oy1; py < ny1; py++)                                          //  top band of pixels
   for (px = ox1; px < ox2; px++) 
   {
      pix1 = PXMpix(E1pxm,px,py);
      pix3 = PXMpix(E3pxm,px,py);
      if (px < trimx1 || px > trimx2 || py < trimy1 || py > trimy2) {
         pix3[0] = pix1[0] / 2;                                            //  outside trim margins
         pix3[1] = pix1[1] / 2;                                            //  50% brightness
         pix3[2] = pix1[2] / 2;
      }
      else {
         pix3[0] = pix1[0];                                                //  inside trim margins
         pix3[1] = pix1[1];                                                //  100% brightness
         pix3[2] = pix1[2];
      }
   }

   for (py = oy1; py < oy2; py++)                                          //  right band
   for (px = nx2; px < ox2; px++)
   {
      pix1 = PXMpix(E1pxm,px,py);
      pix3 = PXMpix(E3pxm,px,py);
      if (px < trimx1 || px > trimx2 || py < trimy1 || py > trimy2) {
         pix3[0] = pix1[0] / 2;
         pix3[1] = pix1[1] / 2;
         pix3[2] = pix1[2] / 2;
      }
      else {
         pix3[0] = pix1[0];
         pix3[1] = pix1[1];
         pix3[2] = pix1[2];
      }
   }

   for (py = ny2; py < oy2; py++)                                          //  bottom band
   for (px = ox1; px < ox2; px++)
   {
      pix1 = PXMpix(E1pxm,px,py);
      pix3 = PXMpix(E3pxm,px,py);
      if (px < trimx1 || px > trimx2 || py < trimy1 || py > trimy2) {
         pix3[0] = pix1[0] / 2;
         pix3[1] = pix1[1] / 2;
         pix3[2] = pix1[2] / 2;
      }
      else {
         pix3[0] = pix1[0];
         pix3[1] = pix1[1];
         pix3[2] = pix1[2];
      }
   }

   for (py = oy1; py < oy2; py++)                                          //  left band
   for (px = ox1; px < nx1; px++)
   {
      pix1 = PXMpix(E1pxm,px,py);
      pix3 = PXMpix(E3pxm,px,py);
      if (px < trimx1 || px > trimx2 || py < trimy1 || py > trimy2) {
         pix3[0] = pix1[0] / 2;
         pix3[1] = pix1[1] / 2;
         pix3[2] = pix1[2] / 2;
      }
      else {
         pix3[0] = pix1[0];
         pix3[1] = pix1[1];
         pix3[2] = pix1[2];
      }
   }
   
   mwpaint3(ox1,oy1,ox2-ox1,ny1-oy1);                                      //  4 updated rectangles
   mwpaint3(nx2,oy1,ox2-nx2,oy2-oy1);
   mwpaint3(ox1,ny2,ox2-ox1,oy2-ny2);
   mwpaint3(ox1,oy1,nx1-ox1,oy2-oy1);

   Ntoplines = 4;                                                          //  outline trim rectangle
   toplinex1[0] = trimx1;
   topliney1[0] = trimy1;
   toplinex2[0] = trimx2;
   topliney2[0] = trimy1;
   toplinex1[1] = trimx2;
   topliney1[1] = trimy1;
   toplinex2[1] = trimx2;
   topliney2[1] = trimy2;
   toplinex1[2] = trimx2;
   topliney1[2] = trimy2;
   toplinex2[2] = trimx1;
   topliney2[2] = trimy2;
   toplinex1[3] = trimx1;
   topliney1[3] = trimy2;
   toplinex2[3] = trimx1;
   topliney2[3] = trimy1;
   draw_toplines(1);

   px = 0.5 * (trimx1 + trimx2);                                           //  add target circle in the middle
   py = 0.5 * (trimy1 + trimy2);                                           //  v.12.09
   radius = 0.02 * (Fpxb->ww + Fpxb->hh);
   radius = radius * Mscale;
   add_topcircle(px,py,radius/2,1);
   add_topcircle(px,py,radius/2+1,2);
   draw_topcircles();
   zmainloop();                                                            //  v.13.03

   return;
}


/**************************************************************************/

//  auto-trim image - set trim rectangle to exclude black margins
//                    left over from rotate or warp functions

void m_autotrim(GtkWidget *, cchar *)
{
   int      px1, py1, px2, py2;
   int      qx1, qy1, qx2, qy2;
   int      qx, qy, step1, step2;
   int      area1, area2, Fgrow = 0;
   float    *ppix;

   if (! E0pxm) return;
   
   px1 = 0.4 * Fpxb->ww;                                                   //  select small rectangle in the middle
   py1 = 0.4 * Fpxb->hh;
   px2 = 0.6 * Fpxb->ww;
   py2 = 0.6 * Fpxb->hh;
   
   step1 = 0.02 * (Fpxb->ww + Fpxb->hh);                                   //  start with big search steps
   step2 = 0.1 * step1;
   
   Ffuncbusy = 1;                                                          //  v.12.09
   
   while (true)
   {
      zmainloop();                                                         //  v.12.09

      while (true)
      {
         Fgrow = 0;

         area1 = (px2 - px1) * (py2 - py1);                                //  area of current selection rectangle
         area2 = area1;

         for (qx1 = px1-step1; qx1 <= px1+step1; qx1 += step2)             //  loop, vary NW and SE corners +/-
         for (qy1 = py1-step1; qy1 <= py1+step1; qy1 += step2) 
         for (qx2 = px2-step1; qx2 <= px2+step1; qx2 += step2) 
         for (qy2 = py2-step1; qy2 <= py2+step1; qy2 += step2) 
         {
            if (qx1 < 0) continue;                                         //  check image limits
            if (qy1 < 0) continue;
            if (qx1 > 0.5 * Fpxb->ww) continue;
            if (qy1 > 0.5 * Fpxb->hh) continue;
            if (qx2 > Fpxb->ww-1) continue;
            if (qy2 > Fpxb->hh-1) continue;
            if (qx2 < 0.5 * Fpxb->ww) continue;
            if (qy2 < 0.5 * Fpxb->hh) continue;
            
            ppix = PXMpix(E0pxm,qx1,qy1);                                  //  check 4 corners are not 
            if (ppix[2] < 1) continue;                                     //    in the black margin zones     v.12.09
            ppix = PXMpix(E0pxm,qx2,qy1);
            if (ppix[2] < 1) continue;
            ppix = PXMpix(E0pxm,qx2,qy2);
            if (ppix[2] < 1) continue;
            ppix = PXMpix(E0pxm,qx1,qy2);
            if (ppix[2] < 1) continue;

            area2 = (qx2 - qx1) * (qy2 - qy1);                             //  look for larger enclosed area
            if (area2 <= area1) continue;
            
            for (qx = qx1; qx < qx2; qx++) {                               //  check top/bottom sides don't intersect
               ppix = PXMpix(E0pxm,qx,qy1);                                //    the black margin zones
               if (ppix[2] < 1) break;                                     //                                  v.12.09
               ppix = PXMpix(E0pxm,qx,qy2);
               if (ppix[2] < 1) break;
            }
            if (qx < qx2) continue;

            for (qy = qy1; qy < qy2; qy++) {                               //  also left/right sides
               ppix = PXMpix(E0pxm,qx1,qy);
               if (ppix[2] < 1) break;
               ppix = PXMpix(E0pxm,qx2,qy);
               if (ppix[2] < 1) break;
            }
            if (qy < qy2) continue;

            Fgrow = 1;                                                     //  successfully grew the rectangle
            px1 = qx1;                                                     //  new bigger rectangle coordinates
            py1 = qy1;                                                     //    for the next search iteration
            px2 = qx2;
            py2 = qy2;
            goto breakout;                                                 //  leave all 4 loops      v.13.08
         }
         
      breakout:
         if (! Fgrow) break;
      }
      
      if (step2 == 1) break;                                               //  done

      step1 = 0.6 * step1;                                                 //  reduce search step size
      step2 = 0.2 * step1;
      if (step2 < 1) step2 = 1;
   }

   trimx1 = px1;                                                           //  set parameters for trim function
   trimx2 = px2;
   trimy1 = py1;
   trimy2 = py2;
   
   Ffuncbusy = 0;                                                          //  v.12.09

   m_trim(0,"autotrim");                                                   //  perform regular trim with
   return;                                                                 //    pre-set margins
}


/**************************************************************************/

//  Resize (rescale) image
//
//  Output pixels are composites of input pixels, e.g. 2/3 size means 
//  that 3x3 input pixels are mapped into 2x2 output pixels, and an 
//  image size of 1000 x 600 becomes 667 x 400.

int      resize_ww0, resize_hh0;                                           //  original size
int      resize_ww1, resize_hh1;                                           //  new size
editfunc    EFresize;

void   resize_initz(zdialog *zd);
int    resize_dialog_event(zdialog *zd, cchar *event);
void * resize_thread(void *);


void m_resize(GtkWidget *, cchar *)
{
   cchar  *lockmess = ZTX("Lock aspect ratio");

   F1_help_topic = "resize";

   EFresize.funcname = "resize";
   EFresize.Frestart = 1;                                                  //  allow restart             v.13.04
   EFresize.threadfunc = resize_thread;                                    //  thread function
   if (! edit_setup(EFresize)) return;                                     //  setup edit

/****
                  pixels     percent
          width   [______]   [______]
          height  [______]   [______]

          presets [2/3] [1/2] [1/3] [1/4] [Prev]

          [_] lock aspect ratio

                            [done] [cancel]
****/

   zdialog *zd = zdialog_new(ZTX("Resize Image"),mWin,Bdone,Bcancel,null);
   EFresize.zd = zd;
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zd,"vbox","vb11","hb1",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb12","hb1",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb13","hb1",0,"homog|space=5");
   zdialog_add_widget(zd,"label","placeholder","vb11",0);
   zdialog_add_widget(zd,"label","labw","vb11",Bwidth);
   zdialog_add_widget(zd,"label","labh","vb11",Bheight);
   zdialog_add_widget(zd,"label","labpix","vb12","pixels");
   zdialog_add_widget(zd,"spin","wpix","vb12","20|9999|1|20");
   zdialog_add_widget(zd,"spin","hpix","vb12","20|9999|1|20");
   zdialog_add_widget(zd,"label","labpct","vb13",Bpercent);
   zdialog_add_widget(zd,"spin","wpct","vb13","1|500|0.1|100");
   zdialog_add_widget(zd,"spin","hpct","vb13","1|500|0.1|100");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","preset","hb2",Bpresets,"space=5");
   zdialog_add_widget(zd,"button","b 3/4","hb2","3/4");
   zdialog_add_widget(zd,"button","b 2/3","hb2","2/3");
   zdialog_add_widget(zd,"button","b 1/2","hb2","1/2");
   zdialog_add_widget(zd,"button","b 1/3","hb2","1/3");
   zdialog_add_widget(zd,"button","b 1/4","hb2","1/4");
   zdialog_add_widget(zd,"button","prev","hb2",Bprev);
   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"check","lock","hb3",lockmess,"space=5");
   
   resize_initz(zd);
   zdialog_run(zd,resize_dialog_event,"save");                             //  run dialog

   return;
}


//  initialization for new image

void resize_initz(zdialog *zd)
{
   resize_ww0 = E1pxm->ww;                                                 //  original width, height
   resize_hh0 = E1pxm->hh;
   zdialog_stuff(zd,"wpix",resize_ww0);
   zdialog_stuff(zd,"hpix",resize_hh0);
   zdialog_stuff(zd,"lock",1);
   return;
}


//  dialog event and completion callback function

int resize_dialog_event(zdialog *zd, cchar *event)
{
   int         lock;
   float       wpct1, hpct1;
   
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) {                                                //  done
         editresize[0] = resize_ww1;                                       //  remember size used 
         editresize[1] = resize_hh1;
         edit_done();
      }
      else edit_cancel();                                                  //  cancel or kill

      Fzoom = 0;
      return 1;
   }
   
   if (strEqu(event,"focus")) return 1;                                    //  ignore focus                 v.13.09

   if (strEqu(event,"save")) {                                             //  from f_save()                v.13.04
      editresize[0] = resize_ww1;                                          //  remember size used 
      editresize[1] = resize_hh1;
      Fzoom = 0;
      signal_thread();                                                     //  do the update
      wait_thread_idle(); 
      return 1;
   }
   
   if (strEqu(event,"restart")) {                                          //  new image file               v.13.04
      resize_initz(zd);
      return 1;
   }
   
   zdialog_fetch(zd,"wpix",resize_ww1);                                    //  get all widget values
   zdialog_fetch(zd,"hpix",resize_hh1);
   zdialog_fetch(zd,"wpct",wpct1);
   zdialog_fetch(zd,"hpct",hpct1);
   zdialog_fetch(zd,"lock",lock);
   
   if (strEqu(event,"b 3/4")) {
      resize_ww1 = (3 * resize_ww0 + 3) / 4;
      resize_hh1 = (3 * resize_hh0 + 3) / 4;
   }

   if (strEqu(event,"b 2/3")) {
      resize_ww1 = (2 * resize_ww0 + 2) / 3;
      resize_hh1 = (2 * resize_hh0 + 2) / 3;
   }

   if (strEqu(event,"b 1/2")) {
      resize_ww1 = (resize_ww0 + 1) / 2;
      resize_hh1 = (resize_hh0 + 1) / 2;
   }

   if (strEqu(event,"b 1/3")) {
      resize_ww1 = (resize_ww0 + 2) / 3;
      resize_hh1 = (resize_hh0 + 2) / 3;
   }

   if (strEqu(event,"b 1/4")) {
      resize_ww1 = (resize_ww0 + 3) / 4;
      resize_hh1 = (resize_hh0 + 3) / 4;
   }

   if (strEqu(event,"prev")) {                                             //  use previous resize values
      resize_ww1 = editresize[0];
      resize_hh1 = editresize[1];
   }

   if (strEqu(event,"wpct"))                                               //  width % - set pixel width
      resize_ww1 = int(wpct1 / 100.0 * resize_ww0 + 0.5);

   if (strEqu(event,"hpct"))                                               //  height % - set pixel height
      resize_hh1 = int(hpct1 / 100.0 * resize_hh0 + 0.5);
   
   if (lock && event[0] == 'w')                                            //  preserve width/height ratio
      resize_hh1 = int(resize_ww1 * (1.0 * resize_hh0 / resize_ww0) + 0.5);
   if (lock && event[0] == 'h') 
      resize_ww1 = int(resize_hh1 * (1.0 * resize_ww0 / resize_hh0) + 0.5);
   
   if (resize_ww1 < 20 || resize_hh1 < 20) return 1;                       //  refuse tiny size       v.12.11

   hpct1 = 100.0 * resize_hh1 / resize_hh0;                                //  set percents to match pixels
   wpct1 = 100.0 * resize_ww1 / resize_ww0;
   
   zdialog_stuff(zd,"wpix",resize_ww1);                                    //  index all widget values
   zdialog_stuff(zd,"hpix",resize_hh1);
   zdialog_stuff(zd,"wpct",wpct1);
   zdialog_stuff(zd,"hpct",hpct1);
   
   signal_thread();                                                        //  update image, no wait for idle

   return 1;
}


//  do the resize job

void * resize_thread(void *)
{
   while (true)
   {
      thread_idle_loop();                                                  //  wait for signal
      
      mutex_lock(&Fpixmap_lock);
      PXM_free(E3pxm);
      E3pxm = PXM_rescale(E1pxm,resize_ww1,resize_hh1);                    //  rescale the edit image
      CEF->Fmods++; 
      CEF->Fsaved = 0;
      mutex_unlock(&Fpixmap_lock);
      mwpaint2();
      zsleep(0.1);                                                         //  v.13.03
      if (! E3pxm) thread_exit();                                          //  memory failure      v.13.03
   }

   return 0;
}


/**************************************************************************/

//  flip an image horizontally or vertically

editfunc    EFflip;

void m_flip(GtkWidget *, cchar *)
{
   int flip_dialog_event(zdialog *zd, cchar *event);

   F1_help_topic = "flip_image";

   EFflip.funcname = "flip";
   EFflip.Frestart = 1;                                                    //  allow restart             v.13.04
   if (! edit_setup(EFflip)) return;                                       //  setup edit

   zdialog *zd = zdialog_new(ZTX("Flip"),mWin,Bdone,Bcancel,null);
   EFflip.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zd,"button","horz","hb1",ZTX("horizontal"),"space=5");
   zdialog_add_widget(zd,"button","vert","hb1",ZTX("vertical"),"space=5");

   zdialog_run(zd,flip_dialog_event,"save");                               //  run dialog, parallel
   return;
}


//  dialog event and completion callback function

int flip_dialog_event(zdialog *zd, cchar *event)
{
   int flip_horz();
   int flip_vert();

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 1;
   }

   if (strEqu(event,"horz")) flip_horz();
   if (strEqu(event,"vert")) flip_vert();
   return 1;
}


int flip_horz()
{
   int      px, py;
   float    *pix3, *pix9;
   
   E9pxm = PXM_copy(E3pxm);
   
   for (py = 0; py < E3pxm->hh; py++)
   for (px = 0; px < E3pxm->ww; px++)
   {
      pix3 = PXMpix(E3pxm,px,py);                                          //  image9 = flipped image3
      pix9 = PXMpix(E9pxm,E3pxm->ww-1-px,py);
      pix9[0] = pix3[0];
      pix9[1] = pix3[1];
      pix9[2] = pix3[2];
   }
   
   mutex_lock(&Fpixmap_lock);
   PXM_free(E3pxm);                                                        //  image9 >> image3
   E3pxm = E9pxm;
   E9pxm = 0;
   mutex_unlock(&Fpixmap_lock);

   CEF->Fmods++; 
   CEF->Fsaved = 0;
   mwpaint2();
   return 0;
}


int flip_vert()
{
   int      px, py;
   float    *pix3, *pix9;
   
   E9pxm = PXM_copy(E3pxm);
   
   for (py = 0; py < E3pxm->hh; py++)
   for (px = 0; px < E3pxm->ww; px++)
   {
      pix3 = PXMpix(E3pxm,px,py);                                          //  image9 = flipped image3
      pix9 = PXMpix(E9pxm,px,E3pxm->hh-1-py);
      pix9[0] = pix3[0];
      pix9[1] = pix3[1];
      pix9[2] = pix3[2];
   }
   
   mutex_lock(&Fpixmap_lock);
   PXM_free(E3pxm);                                                        //  image9 >> image3
   E3pxm = E9pxm;
   E9pxm = 0;
   mutex_unlock(&Fpixmap_lock);

   CEF->Fmods++; 
   CEF->Fsaved = 0;
   mwpaint2();
   return 0;
}


/**************************************************************************/

//  write text on top of the image

namespace writetext 
{
   textattr_t  attr;
   
   char     text[1000] = "text";                                           //  text
   char     file[1000] = "";                                               //  file for write_text data
   int      px, py;                                                        //  text position on image
   int      present;                                                       //  flag, write_text present on image
   PXM      *pxm_text = 0;                                                 //  buffer for rendered text
   PXM      *pxm_transp = 0;                                               //  buffer for transparency data

   int   dialog_event(zdialog *zd, cchar *event);                          //  dialog event function
   void  mousefunc();                                                      //  mouse event function
   void  write(int mode);                                                  //  write text on image

   editfunc    EFwritetext;
}


void m_write_text(GtkWidget *, cchar *menu)
{
   using namespace writetext;

   cchar    *intro = ZTX("Enter text, click/drag on image, right click to remove");

   F1_help_topic = "write_text";                                           //  user guide topic

   EFwritetext.funcname = "write_text";
   EFwritetext.Farea = 1;                                                  //  select area ignored
   EFwritetext.Frestart = 1;                                               //  allow restart                v.13.05
   if (! edit_setup(EFwritetext)) return;                                  //  setup edit

/***
       ___________________________________________________________
      |               Write Text on Image                         |
      |                                                           |
      |  Enter text, click/drag on image, right click to remove.  |
      |                                                           |
      |  Text [_________________________________________]         |
      |                                                           |
      |  [font] [FreeSans         ]  size [ 44|v]                 |
      |                                                           |
      |            color   transp.   width    angle               |
      |  text     [_____]  [___|v]           [___|v]              |
      |  backing  [_____]  [___|v]                                |
      |  outline  [_____]  [___|v]  [___|v]                       |
      |  shadow   [_____]  [___|v]  [___|v]  [___|v]              |
      |                                                           |
      |  Text File: [Open] [Save]                                 |
      |                                                           |
      |                                  [Apply] [Done] [Cancel]  |
      |___________________________________________________________|

***/

   strcpy(attr.font,"FreeSans");                                           //  default font and size
   attr.size = 40;

   zdialog *zd = zdialog_new(ZTX("Write Text on Image"),mWin,Bapply,Bdone,Bcancel,null);
   EFwritetext.zd = zd;
   EFwritetext.mousefunc = mousefunc;
   EFwritetext.menufunc = m_write_text;                                    //  allow restart          v.13.04

   zdialog_add_widget(zd,"label","intro","dialog",intro,"space=3");
   zdialog_add_widget(zd,"hbox","hbtext","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labtext","hbtext",ZTX("Text"),"space=5");
   zdialog_add_widget(zd,"frame","frtext","hbtext",0,"expand");
   zdialog_add_widget(zd,"edit","text","frtext","text","expand");

   zdialog_add_widget(zd,"hbox","hbfont","dialog",0,"space=8");
   zdialog_add_widget(zd,"button","fontbutt","hbfont",Bfont);
   zdialog_add_widget(zd,"entry","fontname","hbfont","FreeSans","space=2|expand");
   zdialog_add_widget(zd,"label","labsize","hbfont",Bsize,"space=2");
   zdialog_add_widget(zd,"spin","fontsize","hbfont","8|500|1|40");
   
   zdialog_add_widget(zd,"hbox","hbcol","dialog");
   zdialog_add_widget(zd,"vbox","vbcol1","hbcol",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vbcol2","hbcol",0,"homog|space=2");
   zdialog_add_widget(zd,"vbox","vbcol3","hbcol",0,"homog|space=2");
   zdialog_add_widget(zd,"vbox","vbcol4","hbcol",0,"homog|space=2");
   zdialog_add_widget(zd,"vbox","vbcol5","hbcol",0,"homog|space=2");

   zdialog_add_widget(zd,"label","space","vbcol1");
   zdialog_add_widget(zd,"label","labtext","vbcol1",ZTX("text"));
   zdialog_add_widget(zd,"label","labback","vbcol1",ZTX("backing"));
   zdialog_add_widget(zd,"label","laboutln","vbcol1",ZTX("outline"));
   zdialog_add_widget(zd,"label","labshadow","vbcol1",ZTX("shadow"));

   zdialog_add_widget(zd,"label","labcol","vbcol2",ZTX("Color"));
   zdialog_add_widget(zd,"colorbutt","fgcolor","vbcol2","0|0|0");
   zdialog_add_widget(zd,"colorbutt","bgcolor","vbcol2","255|255|255");
   zdialog_add_widget(zd,"colorbutt","tocolor","vbcol2","255|0|0");
   zdialog_add_widget(zd,"colorbutt","shcolor","vbcol2","255|0|0");

   zdialog_add_widget(zd,"label","labcol","vbcol3",ZTX("Transparency"));
   zdialog_add_widget(zd,"spin","fgtransp","vbcol3","0|100|1|0");
   zdialog_add_widget(zd,"spin","bgtransp","vbcol3","0|100|1|0");
   zdialog_add_widget(zd,"spin","totransp","vbcol3","0|100|1|0");
   zdialog_add_widget(zd,"spin","shtransp","vbcol3","0|100|1|0");

   zdialog_add_widget(zd,"label","labw","vbcol4",ZTX("Width"));
   zdialog_add_widget(zd,"label","space","vbcol4");
   zdialog_add_widget(zd,"label","space","vbcol4");
   zdialog_add_widget(zd,"spin","towidth","vbcol4","0|30|1|0");
   zdialog_add_widget(zd,"spin","shwidth","vbcol4","0|50|1|0");

   zdialog_add_widget(zd,"label","labw","vbcol5",Bangle);
   zdialog_add_widget(zd,"spin","angle","vbcol5","-180|180|0.5|0");
   zdialog_add_widget(zd,"label","space","vbcol5");
   zdialog_add_widget(zd,"label","space","vbcol5");
   zdialog_add_widget(zd,"spin","shangle","vbcol5","-180|180|1|0");

   zdialog_add_widget(zd,"hbox","hbaf","dialog",0,"space=10");
   zdialog_add_widget(zd,"label","labbg","hbaf",ZTX("Text File:"),"space=3");
   zdialog_add_widget(zd,"button","loadtext","hbaf",Bopen,"space=5");
   zdialog_add_widget(zd,"button","savetext","hbaf",Bsave,"space=5");

   zdialog_restore_inputs(zd);                                             //  restore prior inputs
   zdialog_run(zd,dialog_event,"save");                                    //  run dialog, parallel

   zdialog_fetch(zd,"text",text,1000);                                     //  get defaults or prior inputs
   zdialog_fetch(zd,"fontname",attr.font,100);
   zdialog_fetch(zd,"fontsize",attr.size);
   zdialog_fetch(zd,"angle",attr.angle);
   zdialog_fetch(zd,"fgcolor",attr.color[0],20);
   zdialog_fetch(zd,"bgcolor",attr.color[1],20);
   zdialog_fetch(zd,"tocolor",attr.color[2],20);
   zdialog_fetch(zd,"shcolor",attr.color[3],20);
   zdialog_fetch(zd,"fgtransp",attr.transp[0]);
   zdialog_fetch(zd,"bgtransp",attr.transp[1]);
   zdialog_fetch(zd,"totransp",attr.transp[2]);
   zdialog_fetch(zd,"shtransp",attr.transp[3]);
   zdialog_fetch(zd,"towidth",attr.towidth);
   zdialog_fetch(zd,"shwidth",attr.shwidth);
   zdialog_fetch(zd,"shangle",attr.shangle);

   gentext(text,&attr,pxm_text,pxm_transp);                                //  initial text

   takeMouse(mousefunc,dragcursor);                                        //  connect mouse function
   present = 0;                                                            //  no text on image yet
   px = py = -1;                                                           //  no position defined yet

   return;
}


//  dialog event and completion callback function

int writetext::dialog_event(zdialog *zd, cchar *event)
{
   using namespace writetext;

   GtkWidget   *font_dialog;
   char        font[100];
   int         size;
   char        *pp;
   
   if (zd->zstat)
   {
      if (zd->zstat == 1) {                                                //  Apply, commit present text to image
         write(1);                                                         //  v.12.03
         present = 0;                                                      //  (no old text to erase)
         px = py = -1;
         zd->zstat = 0;
         edit_commit();                                                    //  v.13.07
         return 1;
      }

      if (zd->zstat == 2 && CEF->Fmods)                                    //  Done, complete pending edit
         edit_done();
      else edit_cancel();                                                  //  Cancel or kill
      PXM_free(pxm_text);                                                  //  free memory                  v.12.10
      PXM_free(pxm_transp);
      return 1;
   }
   
   if (strEqu(event,"save")) {                                             //  from f_save()                v.13.04
      write(1);
      present = 0;
      px = py = -1;
      return 1;
   }
   
   if (strEqu(event,"restart")) {                                          //  new image                    v.13.04
      present = 0;
      px = py = -1;
      return 1;
   }

   if (strEqu(event,"focus")) {                                            //  toggle mouse capture
      takeMouse(mousefunc,dragcursor);                                     //  connect mouse function
      return 1;
   }

   if (strEqu(event,"loadtext"))                                           //  load zdialog fields from file
   {
      load_text(zd);
      zdialog_fetch(zd,"text",text,1000);                                  //  get all zdialog fields
      zdialog_fetch(zd,"fontname",attr.font,80);
      zdialog_fetch(zd,"fontsize",attr.size);
      zdialog_fetch(zd,"angle",attr.angle);
      zdialog_fetch(zd,"fgcolor",attr.color[0],20);
      zdialog_fetch(zd,"bgcolor",attr.color[1],20);
      zdialog_fetch(zd,"tocolor",attr.color[2],20);
      zdialog_fetch(zd,"shcolor",attr.color[3],20);
      zdialog_fetch(zd,"fgtransp",attr.transp[0]);
      zdialog_fetch(zd,"bgtransp",attr.transp[1]);
      zdialog_fetch(zd,"totransp",attr.transp[2]);
      zdialog_fetch(zd,"shtransp",attr.transp[3]);
      zdialog_fetch(zd,"towidth",attr.towidth);
      zdialog_fetch(zd,"shwidth",attr.shwidth);
      zdialog_fetch(zd,"shangle",attr.shangle);
   }

   if (strEqu(event,"savetext")) {                                         //  save zdialog fields to file
      save_text(zd);
      return 1;
   }

   if (strEqu(event,"text"))                                               //  get text from dialog
      zdialog_fetch(zd,"text",text,1000);

   if (strEqu(event,"fontbutt")) {                                         //  select new font
      snprintf(font,100,"%s %d",attr.font,attr.size);
      font_dialog = gtk_font_chooser_dialog_new(ZTX("select font"),MWIN);
      gtk_font_chooser_set_font(GTK_FONT_CHOOSER(font_dialog),font);
      gtk_dialog_run(GTK_DIALOG(font_dialog));
      pp = gtk_font_chooser_get_font(GTK_FONT_CHOOSER(font_dialog));
      gtk_widget_destroy(font_dialog);

      if (pp) {                                                            //  should have "fontname nn"
         strncpy0(font,pp,100);
         g_free(pp);
         pp = font + strlen(font);
         while (*pp != ' ') pp--;
         if (pp > font) {
            size = atoi(pp);
            if (size < 8) size = 8;
            zdialog_stuff(zd,"fontsize",size);
            attr.size = size;
            *pp = 0;
            strcpy(attr.font,font);                                        //  get fontname = new font name
            zdialog_stuff(zd,"fontname",font);
         }
      }
   }

   if (strEqu(event,"fontsize"))                                           //  new font size
      zdialog_fetch(zd,"fontsize",attr.size);

   if (strEqu(event,"angle"))
      zdialog_fetch(zd,"angle",attr.angle);
   
   if (strEqu(event,"fgcolor"))                                            //  foreground (text) color
      zdialog_fetch(zd,"fgcolor",attr.color[0],20);
   
   if (strEqu(event,"bgcolor"))                                            //  background color
      zdialog_fetch(zd,"bgcolor",attr.color[1],20);
   
   if (strEqu(event,"tocolor"))                                            //  text outline color
      zdialog_fetch(zd,"tocolor",attr.color[2],20);
   
   if (strEqu(event,"shcolor"))                                            //  text shadow color
      zdialog_fetch(zd,"shcolor",attr.color[3],20);
   
   if (strEqu(event,"fgtransp"))                                           //  foreground transparency
      zdialog_fetch(zd,"fgtransp",attr.transp[0]);
   
   if (strEqu(event,"bgtransp"))                                           //  background transparency
      zdialog_fetch(zd,"bgtransp",attr.transp[1]);

   if (strEqu(event,"totransp"))                                           //  text outline transparency
      zdialog_fetch(zd,"totransp",attr.transp[2]);
   
   if (strEqu(event,"shtransp"))                                           //  text shadow transparency
      zdialog_fetch(zd,"shtransp",attr.transp[3]);
   
   if (strEqu(event,"towidth"))                                            //  text outline width
      zdialog_fetch(zd,"towidth",attr.towidth);

   if (strEqu(event,"shwidth"))                                            //  text shadow width
      zdialog_fetch(zd,"shwidth",attr.shwidth);

   if (strEqu(event,"shangle"))                                            //  text shadow angle
      zdialog_fetch(zd,"shangle",attr.shangle);

   gentext(text,&attr,pxm_text,pxm_transp);                                //  build text image from text and attributes
   write(1);                                                               //  write on image
   zmainloop();

   return 1;
}


//  mouse function, set position for text on image

void writetext::mousefunc()
{
   using namespace writetext;

   if (LMclick) {                                                          //  left mouse click
      px = Mxclick;                                                        //  new text position on image
      py = Myclick;
      write(1);                                                            //  erase old, write new text on image
   }
   
   if (RMclick) {                                                          //  right mouse click
      write(2);                                                            //  erase old text, if any
      px = py = -1;                                                        //  v.12.03
   }
   
   if (Mxdrag || Mydrag)                                                   //  mouse dragged
   {
      px = Mxdrag;                                                         //  new text position on image
      py = Mydrag;
      write(1);                                                            //  erase old, write new text on image
   }
   
   LMclick = RMclick = Mxdrag = Mydrag = 0;                                //  v.12.11
   return;
}


//  write text on image at designated location
//  mode: 1  erase old and write to new position
//        2  erase old and write nothing

void writetext::write(int mode)                                            //  v.12.02
{
   using namespace writetext;

   float       *pix1, *pix2, *pix3;
   int         px1, py1, px3, py3, done;
   float       e3part;

   static int  orgx1, orgy1, ww1, hh1;                                     //  old text image overlap rectangle
   int         orgx2, orgy2, ww2, hh2;                                     //  new overlap rectangle
   
   if (present)                                                            //  v.13.03
   {
      for (py3 = orgy1; py3 < orgy1 + hh1; py3++)                          //  erase prior text image
      for (px3 = orgx1; px3 < orgx1 + ww1; px3++)                          //  replace E3 pixels with E1 pixels
      {                                                                    //    in prior overlap rectangle
         if (px3 < 0 || px3 >= E3pxm->ww) continue;
         if (py3 < 0 || py3 >= E3pxm->hh) continue;
         pix1 = PXMpix(E1pxm,px3,py3);
         pix3 = PXMpix(E3pxm,px3,py3);
         pix3[0] = pix1[0];
         pix3[1] = pix1[1];
         pix3[2] = pix1[2];
      }
   }
   
   done = 0;
   if (mode == 2) done = 1;                                                //  erase only
   if (! *text) done = 1;                                                  //  no text defined
   if (px < 0 && py < 0) done = 1;                                         //  no position defined
   
   if (done) {
      if (present) {
         mwpaint3(orgx1,orgy1,ww1,hh1);                                    //  update window to erase old text
         present = 0;                                                      //  mark no text present
         CEF->Fmods--;                                                     //  reduce mod count
         CEF->Fsaved = 0;
      }
      return;
   }

   ww2 = pxm_text->ww;                                                     //  text image buffer
   hh2 = pxm_text->hh;

   orgx2 = px - ww2/2;                                                     //  copy-to image3 location
   orgy2 = py - hh2/2;

   for (py1 = 0; py1 < hh2; py1++)                                         //  loop all pixels in text image
   for (px1 = 0; px1 < ww2; px1++)
   {
      px3 = orgx2 + px1;                                                   //  copy-to image3 pixel
      py3 = orgy2 + py1;

      if (px3 < 0 || px3 >= E3pxm->ww) continue;                           //  omit parts beyond edges
      if (py3 < 0 || py3 >= E3pxm->hh) continue;

      pix1 = PXMpix(pxm_text,px1,py1);                                     //  copy-from text pixel
      pix2 = PXMpix(pxm_transp,px1,py1);                                   //  copy-from transparency
      pix3 = PXMpix(E3pxm,px3,py3);                                        //  copy-to image pixel
      
      e3part = pix2[0] / 256.0;                                            //  image part visible through text
      
      pix3[0] = pix1[0] + e3part * pix3[0];                                //  combine text part + image part
      pix3[1] = pix1[1] + e3part * pix3[1];
      pix3[2] = pix1[2] + e3part * pix3[2];
   }
   
   if (present) {
      mwpaint3(orgx1,orgy1,ww1,hh1);                                       //  update window to erase old text
      CEF->Fmods--;                                                        //  reduce mod count             v.12.02
      CEF->Fsaved = 0;
      present = 0;
   }
                                                                           //  (updates together to reduce flicker)
   mwpaint3(orgx2,orgy2,ww2,hh2);                                          //  update window for new text

   CEF->Fmods++;                                                           //  increase mod count           v.12.02
   CEF->Fsaved = 0;
   present = 1;                                                            //  mark text is present

   orgx1 = orgx2;                                                          //  remember overlap rectangle
   orgy1 = orgy2;                                                          //    for next call
   ww1 = ww2;
   hh1 = hh2;
   return;
}


/**************************************************************************/

//  Create a graphic image with text, white on black, any font, any angle.
//  Add outline and shadow colors. Apply transparencies.

int gentext(cchar *text, textattr_t *attr, PXM *&pxm_text, PXM *&pxm_transp)        //  v.13.09
{
   PXM * gentext_outline(textattr_t *, PXM *);
   PXM * gentext_shadow(textattr_t *, PXM *);

   PangoFontDescription    *pfont;
   PangoLayout             *playout;
   cairo_surface_t         *surface;
   cairo_t                 *cr;

   float          angle;
   int            fgred, fggreen, fgblue;
   int            bgred, bggreen, bgblue;
   int            tored, togreen, toblue;
   int            shred, shgreen, shblue;
   float          fgtransp, bgtransp, totransp, shtransp;

   PXM            *pxm_temp1, *pxm_temp2, *pxm_temp3;
   uint8          *cairo_data, *cpix;
   float          *pix1, *pix2, *pix3;
   int            px, py, ww, hh;
   char           font[100];
   int            red, green, blue;
   float          fgpart, topart, shpart, bgpart;
   
   zthreadcrash();                                                         //  thread usage not allowed

   if (! *text) text = " ";

   snprintf(font,100,"%s %d",attr->font,attr->size);                       //  fontname nn
   angle = attr->angle;
   fgred = atoi(strField(attr->color[0],'|',1));                           //  get text foreground color
   fggreen = atoi(strField(attr->color[0],'|',2));
   fgblue = atoi(strField(attr->color[0],'|',3));
   bgred = atoi(strField(attr->color[1],'|',1));                           //  get text background color
   bggreen = atoi(strField(attr->color[1],'|',2));
   bgblue = atoi(strField(attr->color[1],'|',3));
   tored = atoi(strField(attr->color[2],'|',1));                           //  get text outline color
   togreen = atoi(strField(attr->color[2],'|',2));
   toblue = atoi(strField(attr->color[2],'|',3));
   shred = atoi(strField(attr->color[3],'|',1));                           //  get text shadow color
   shgreen = atoi(strField(attr->color[3],'|',2));
   shblue = atoi(strField(attr->color[3],'|',3));
   fgtransp = 0.01 * attr->transp[0];                                      //  get transparencies
   bgtransp = 0.01 * attr->transp[1];                                      //  text, background, outline, shadow
   totransp = 0.01 * attr->transp[2];
   shtransp = 0.01 * attr->transp[3];

   pfont = pango_font_description_from_string(font);                       //  make layout with text
   playout = gtk_widget_create_pango_layout(cDwin,null);
   pango_layout_set_font_description(playout,pfont);
   pango_layout_set_text(playout,text,-1);

   pango_layout_get_pixel_size(playout,&ww,&hh);
   ww += 2 + 0.2 * attr->size;                                             //  compensate bad font metrics
   hh += 2 + 0.1 * attr->size;
   
   surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,ww,hh);         //  cairo output image
   cr = cairo_create(surface);
   pango_cairo_show_layout(cr,playout);                                    //  write text layout to image
   cairo_data = cairo_image_surface_get_data(surface);                     //  output text image pixels
   
   pxm_temp1 = PXM_make(ww,hh);                                            //  create PXM

   for (py = 0; py < hh; py++)                                             //  copy text image to PXM
   for (px = 0; px < ww; px++)
   {
      cpix = cairo_data + 4 * (ww * py + px);                              //  pango output is monocolor
      pix2 = PXMpix(pxm_temp1,px,py);
      pix2[0] = cpix[3];
      pix2[1] = pix2[2] = 0;
   }

   pango_font_description_free(pfont);                                     //  free resources
   g_object_unref(playout);
   cairo_destroy(cr);
   cairo_surface_destroy(surface);

   pxm_temp2 = gentext_outline(attr,pxm_temp1);                            //  add text outline color if any
   if (pxm_temp2) {
      PXM_free(pxm_temp1);
      pxm_temp1 = pxm_temp2;
   }
   
   pxm_temp2 = gentext_shadow(attr,pxm_temp1);                             //  add text shadow color if any
   if (pxm_temp2) {
      PXM_free(pxm_temp1);
      pxm_temp1 = pxm_temp2;
   }
   
   if (fabsf(angle) > 0.1) {                                               //  rotate text if wanted
      pxm_temp2 = PXM_rotate(pxm_temp1,angle);
      PXM_free(pxm_temp1);
      pxm_temp1 = pxm_temp2;
   }

   ww = pxm_temp1->ww;                                                     //  text image input PXM
   hh = pxm_temp1->hh;
   pxm_temp2 = PXM_make(ww,hh);                                            //  text image output PXM
   pxm_temp3 = PXM_make(ww,hh);                                            //  output transparency map

   for (py = 0; py < hh; py++)                                             //  loop all pixels in text image
   for (px = 0; px < ww; px++)
   {
      pix1 = PXMpix(pxm_temp1,px,py);                                      //  copy-from pixel (text + outline + shadow)
      pix2 = PXMpix(pxm_temp2,px,py);                                      //  copy-to pixel
      pix3 = PXMpix(pxm_temp3,px,py);                                      //  copy-to transparency
      
      fgpart = pix1[0] / 256.0;
      topart = pix1[1] / 256.0;
      shpart = pix1[2] / 256.0;
      bgpart = (1.0 - fgpart - topart - shpart);
      fgpart = fgpart * (1.0 - fgtransp);
      topart = topart * (1.0 - totransp);
      shpart = shpart * (1.0 - shtransp);
      bgpart = bgpart * (1.0 - bgtransp);
      
      red = fgpart * fgred + topart * tored + shpart * shred + bgpart * bgred;
      green = fgpart * fggreen + topart * togreen + shpart * shgreen + bgpart * bggreen;
      blue = fgpart * fgblue + topart * toblue + shpart * shblue + bgpart * bgblue;

      pix2[0] = red;                                                       //  output total red, green blue
      pix2[1] = green;
      pix2[2] = blue;

      pix3[0] = 255 * (1.0 - fgpart - topart - shpart - bgpart);           //  image part visible through text
   }

   PXM_free(pxm_temp1);
   pxm_text = pxm_temp2;
   pxm_transp = pxm_temp3;
   return 0;
}


//  add an outline color to the text character edges
//  red color [0] is original monocolor text
//  use green color [1] for added outline

PXM * gentext_outline(textattr_t *attr, PXM *pxm1)                         //  v.13.09
{
   PXM         *pxm2;
   int         toww, ww1, hh1, ww2, hh2;
   int         px, py, dx, dy;
   float       *pix1, *pix2, theta;

   toww = attr->towidth;                                                   //  text outline color width
   if (toww == 0) return 0;                                                //  zero

   ww1 = pxm1->ww;                                                         //  input PXM dimensions
   hh1 = pxm1->hh;
   ww2 = ww1 + toww * 2;                                                   //  add margins for outline width
   hh2 = hh1 + toww * 2;
   pxm2 = PXM_make(ww2,hh2);                                               //  output PXM
   
   for (py = 0; py < hh1; py++)                                            //  copy text pixels to outline pixels
   for (px = 0; px < ww1; px++)                                            //    displaced by outline width
   {
      pix1 = PXMpix(pxm1,px,py);
      pix2 = PXMpix(pxm2,px+toww,py+toww);
      pix2[0] = pix1[0];
   }
   
   theta = 0.7 / toww;
   for (theta = 0; theta < 6.3; theta += 0.7/toww)                         //  displace outline pixels in all directions
   {
      dx = roundf(toww * sinf(theta));
      dy = roundf(toww * cosf(theta));

      for (py = 0; py < hh1; py++)
      for (px = 0; px < ww1; px++)
      {
         pix1 = PXMpix(pxm1,px,py);
         pix2 = PXMpix(pxm2,px+toww+dx,py+toww+dy);
         if (pix2[1] < pix1[0] - pix2[0])                                  //  compare text to outline brightness 
            pix2[1] = pix1[0] - pix2[0];                                   //  brighter part is outline pixel
      }
   }
   
   return pxm2;                                                            //  pix2[0] / pix2[1] = text / outline
}


//  add a shadow to the text character edges
//  red color [0] is original monocolor text
//  green color [1] is added outline if any
//  use blue color [2] for added shadow

PXM * gentext_shadow(textattr_t *attr, PXM *pxm1)                          //  v.13.09
{
   PXM         *pxm2;
   int         shww, ww1, hh1, ww2, hh2;
   int         px, py, dx, dy;
   float       *pix1, *pix2, theta;

   shww = attr->shwidth;                                                   //  text shadow width
   if (shww == 0) return 0;                                                //  zero

   ww1 = pxm1->ww;                                                         //  input PXM dimensions
   hh1 = pxm1->hh;
   ww2 = ww1 + shww * 2;                                                   //  add margins for shadow width
   hh2 = hh1 + shww * 2;
   pxm2 = PXM_make(ww2,hh2);                                               //  output PXM
   
   for (py = 0; py < hh1; py++)                                            //  copy text pixels to shadow pixels
   for (px = 0; px < ww1; px++)                                            //    displaced by shadow width
   {
      pix1 = PXMpix(pxm1,px,py);
      pix2 = PXMpix(pxm2,px+shww,py+shww);
      pix2[0] = pix1[0];
      pix2[1] = pix1[1];
   }
   
   theta = (90 - attr->shangle) / 57.3;                                    //  degrees to radians, 0 = to the right
   dx = roundf(shww * sinf(theta));
   dy = roundf(shww * cosf(theta));

   for (py = 0; py < hh1; py++)                                            //  displace text by shadow width
   for (px = 0; px < ww1; px++)
   {
      pix1 = PXMpix(pxm1,px,py);
      pix2 = PXMpix(pxm2,px+shww+dx,py+shww+dy);
      if (pix2[2] < pix1[0] + pix1[1] - pix2[0] - pix2[1])                 //  compare text+outline to shadow pixels
         pix2[2] = pix1[0] + pix1[1] - pix2[0] - pix2[1];                  //  brighter part is shadow pixel
   }
   
   return pxm2;                                                            //  pix2[0] / pix2[1] / pix2[2]
}                                                                          //    = text / outline / shadow brightness


//  load text and attributes from a file

void load_text(zdialog *zd)                                                //  v.13.09
{
   FILE        *fid;
   int         err, nn;
   char        *pp, *pp2, *file, buff[1200];
   cchar       *dialogtitle = "load text data from a file";
   char        text[1000];                                                 //  text data
   textattr_t  attr;

   file = zgetfile1(dialogtitle,"open",writetext_dirk);                    //  get input file from user
   if (! file) return;
   
   fid = fopen(file,"r");                                                  //  open for read
   if (! fid) {
      zmessageACK(mWin,0,"%s",strerror(errno));
      zfree(file);
      return;
   }

   pp = fgets_trim(buff,1200,fid);                                         //  read text string
   if (! pp) goto badfile;
   if (strnNeq(pp,"text string: ",13)) goto badfile;
   pp += 13;
   if (strlen(pp) < 2) goto badfile;
   repl_Nstrs(pp,text,"\\n","\n",null);                                    //  replace "\n" with newline char.

   pp = fgets_trim(buff,1200,fid);                                         //  read font and size
   if (! pp) goto badfile;
   if (strnNeq(pp,"font: ",6)) goto badfile;
   pp += 6;
   strTrim(pp);
   pp2 = strstr(pp,"size: ");
   if (! pp2) goto badfile;
   *pp2 = 0;
   pp2 += 6;
   strncpy0(attr.font,pp,80);
   attr.size = atoi(pp2);
   if (attr.size < 8) goto badfile;

   pp = fgets_trim(buff,1200,fid);                                         //  read text attributes
   if (! pp) goto badfile;
   nn = sscanf(pp,"attributes: %f  %s %s %s %s  %d %d %d %d  %d %d %d",
            &attr.angle, attr.color[0], attr.color[1], attr.color[2], attr.color[3],
            &attr.transp[0], &attr.transp[1], &attr.transp[2], &attr.transp[3],
            &attr.towidth, &attr.shwidth, &attr.shangle);
   if (nn != 12) goto badfile;
   
   err = fclose(fid);
   if (err) {
      zmessageACK(mWin,0,"%s",strerror(errno));
      zfree(file);
      return;
   }
   
   zdialog_stuff(zd,"text",text);                                          //  stuff zdialog fields
   zdialog_stuff(zd,"fontname",attr.font);
   zdialog_stuff(zd,"fontsize",attr.size);
   zdialog_stuff(zd,"angle",attr.angle);
   zdialog_stuff(zd,"fgcolor",attr.color[0]);
   zdialog_stuff(zd,"bgcolor",attr.color[1]);
   zdialog_stuff(zd,"tocolor",attr.color[2]);
   zdialog_stuff(zd,"shcolor",attr.color[3]);
   zdialog_stuff(zd,"fgtransp",attr.transp[0]);
   zdialog_stuff(zd,"bgtransp",attr.transp[1]);
   zdialog_stuff(zd,"totransp",attr.transp[2]);
   zdialog_stuff(zd,"shtransp",attr.transp[3]);
   zdialog_stuff(zd,"towidth",attr.towidth);
   zdialog_stuff(zd,"shwidth",attr.shwidth);
   zdialog_stuff(zd,"shangle",attr.shangle);
   return;

badfile:                                                                   //  project file had a problem
   fclose(fid);
   zmessageACK(mWin,0,ZTX("text file is defective"));
   printf("buff: %s\n",buff);
}


//  save text to a file

void save_text(zdialog *zd)                                                //  v.13.09
{
   cchar       *dialogtitle = "save text data to a file";
   FILE        *fid;
   char        *file, text[1000], text2[1200];
   textattr_t  attr;

   file = zgetfile1(dialogtitle,"save",writetext_dirk);                    //  get output file from user
   if (! file) return;
   
   fid = fopen(file,"w");                                                  //  open for write
   if (! fid) {
      zmessageACK(mWin,0,"%s",strerror(errno));
      zfree(file);
      return;
   }

   zdialog_fetch(zd,"text",text,1000);                                     //  get text and attributes from zdialog
   zdialog_fetch(zd,"fontname",attr.font,80);
   zdialog_fetch(zd,"fontsize",attr.size);
   zdialog_fetch(zd,"angle",attr.angle);
   zdialog_fetch(zd,"fgcolor",attr.color[0],20);
   zdialog_fetch(zd,"bgcolor",attr.color[1],20);
   zdialog_fetch(zd,"tocolor",attr.color[2],20);
   zdialog_fetch(zd,"shcolor",attr.color[3],20);
   zdialog_fetch(zd,"fgtransp",attr.transp[0]);
   zdialog_fetch(zd,"bgtransp",attr.transp[1]);
   zdialog_fetch(zd,"totransp",attr.transp[2]);
   zdialog_fetch(zd,"shtransp",attr.transp[3]);
   zdialog_fetch(zd,"towidth",attr.towidth);
   zdialog_fetch(zd,"shwidth",attr.shwidth);
   zdialog_fetch(zd,"shangle",attr.shangle);

   repl_Nstrs(text, text2,"\n","\\n",null);                                //  replace newlines with "\n"

   fprintf(fid,"text string: %s\n",text2);
   
   fprintf(fid,"font: %s  size: %d \n",attr.font, attr.size);

   fprintf(fid,"attributes: %.4f  %s %s %s %s  %d %d %d %d  %d %d %d \n", 
            attr.angle, attr.color[0], attr.color[1], attr.color[2], attr.color[3],
            attr.transp[0], attr.transp[1], attr.transp[2], attr.transp[3],
            attr.towidth, attr.shwidth, attr.shangle);

   fclose(fid);
   return;
}



